# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .int64 import Int64
from .liquidity_pool_constant_product_parameters import (
    LiquidityPoolConstantProductParameters,
)

__all__ = ["LiquidityPoolEntryConstantProduct"]


class LiquidityPoolEntryConstantProduct:
    """
    XDR Source Code::

        struct
                {
                    LiquidityPoolConstantProductParameters params;

                    int64 reserveA;        // amount of A in the pool
                    int64 reserveB;        // amount of B in the pool
                    int64 totalPoolShares; // total number of pool shares issued
                    int64 poolSharesTrustLineCount; // number of trust lines for the
                                                    // associated pool shares
                }
    """

    def __init__(
        self,
        params: LiquidityPoolConstantProductParameters,
        reserve_a: Int64,
        reserve_b: Int64,
        total_pool_shares: Int64,
        pool_shares_trust_line_count: Int64,
    ) -> None:
        self.params = params
        self.reserve_a = reserve_a
        self.reserve_b = reserve_b
        self.total_pool_shares = total_pool_shares
        self.pool_shares_trust_line_count = pool_shares_trust_line_count

    def pack(self, packer: Packer) -> None:
        self.params.pack(packer)
        self.reserve_a.pack(packer)
        self.reserve_b.pack(packer)
        self.total_pool_shares.pack(packer)
        self.pool_shares_trust_line_count.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LiquidityPoolEntryConstantProduct:
        params = LiquidityPoolConstantProductParameters.unpack(unpacker)
        reserve_a = Int64.unpack(unpacker)
        reserve_b = Int64.unpack(unpacker)
        total_pool_shares = Int64.unpack(unpacker)
        pool_shares_trust_line_count = Int64.unpack(unpacker)
        return cls(
            params=params,
            reserve_a=reserve_a,
            reserve_b=reserve_b,
            total_pool_shares=total_pool_shares,
            pool_shares_trust_line_count=pool_shares_trust_line_count,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LiquidityPoolEntryConstantProduct:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LiquidityPoolEntryConstantProduct:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.params,
                self.reserve_a,
                self.reserve_b,
                self.total_pool_shares,
                self.pool_shares_trust_line_count,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.params == other.params
            and self.reserve_a == other.reserve_a
            and self.reserve_b == other.reserve_b
            and self.total_pool_shares == other.total_pool_shares
            and self.pool_shares_trust_line_count == other.pool_shares_trust_line_count
        )

    def __str__(self):
        out = [
            f"params={self.params}",
            f"reserve_a={self.reserve_a}",
            f"reserve_b={self.reserve_b}",
            f"total_pool_shares={self.total_pool_shares}",
            f"pool_shares_trust_line_count={self.pool_shares_trust_line_count}",
        ]
        return f"<LiquidityPoolEntryConstantProduct [{', '.join(out)}]>"



from decimal import Decimal
from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..asset import Asset
from ..muxed_account import MuxedAccount
from ..utils import raise_if_not_valid_amount
from .operation import Operation

__all__ = ["Clawback"]


class Clawback(Operation):
    """The :class:`Clawback` object, which represents a Clawback operation on
    Stellar's network.

    Claws back an amount of an asset from an account.

    Threshold: Medium

    See `Clawback <https://developers.stellar.org/docs/start/list-of-operations/#clawback>`_ for more information.

    :param asset: The asset being clawed back.
    :param from_: The public key of the account to claw back from.
    :param amount: The amount of the asset to claw back.
    :param source: The source account for the operation. Defaults to the
        transaction's source account.
    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = stellar_xdr.OperationType.CLAWBACK

    def __init__(
        self,
        asset: Asset,
        from_: Union[MuxedAccount, str],
        amount: Union[str, Decimal],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> None:
        super().__init__(source)
        if isinstance(from_, str):
            from_ = MuxedAccount.from_account(from_)
        self.from_: MuxedAccount = from_
        self.asset: Asset = asset
        self.amount: str = str(amount)
        raise_if_not_valid_amount(self.amount, "amount")

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        asset = self.asset.to_xdr_object()
        from_ = self.from_.to_xdr_object()
        amount = Operation.to_xdr_amount(self.amount)
        clawback_op = stellar_xdr.ClawbackOp(asset, from_, stellar_xdr.Int64(amount))
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE, clawback_op=clawback_op
        )
        return body

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Operation) -> "Clawback":
        """Creates a :class:`Clawback` object from an XDR Operation
        object.
        """
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.clawback_op is not None
        asset = Asset.from_xdr_object(xdr_object.body.clawback_op.asset)
        from_ = MuxedAccount.from_xdr_object(xdr_object.body.clawback_op.from_)
        amount = Operation.from_xdr_amount(xdr_object.body.clawback_op.amount.int64)

        op = cls(source=source, from_=from_, asset=asset, amount=amount)
        return op

    def __str__(self):
        return f"<Clawback [asset={self.asset}, from_={self.from_}, amount={self.amount}, source={self.source}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .extension_point import ExtensionPoint
from .uint32 import Uint32

__all__ = ["ExtendFootprintTTLOp"]


class ExtendFootprintTTLOp:
    """
    XDR Source Code::

        struct ExtendFootprintTTLOp
        {
            ExtensionPoint ext;
            uint32 extendTo;
        };
    """

    def __init__(
        self,
        ext: ExtensionPoint,
        extend_to: Uint32,
    ) -> None:
        self.ext = ext
        self.extend_to = extend_to

    def pack(self, packer: Packer) -> None:
        self.ext.pack(packer)
        self.extend_to.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ExtendFootprintTTLOp:
        ext = ExtensionPoint.unpack(unpacker)
        extend_to = Uint32.unpack(unpacker)
        return cls(
            ext=ext,
            extend_to=extend_to,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ExtendFootprintTTLOp:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ExtendFootprintTTLOp:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ext,
                self.extend_to,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.ext == other.ext and self.extend_to == other.extend_to

    def __str__(self):
        out = [
            f"ext={self.ext}",
            f"extend_to={self.extend_to}",
        ]
        return f"<ExtendFootprintTTLOp [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .asset import Asset
from .claimant import Claimant
from .int64 import Int64

__all__ = ["CreateClaimableBalanceOp"]


class CreateClaimableBalanceOp:
    """
    XDR Source Code::

        struct CreateClaimableBalanceOp
        {
            Asset asset;
            int64 amount;
            Claimant claimants<10>;
        };
    """

    def __init__(
        self,
        asset: Asset,
        amount: Int64,
        claimants: List[Claimant],
    ) -> None:
        _expect_max_length = 10
        if claimants and len(claimants) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `claimants` should be {_expect_max_length}, but got {len(claimants)}."
            )
        self.asset = asset
        self.amount = amount
        self.claimants = claimants

    def pack(self, packer: Packer) -> None:
        self.asset.pack(packer)
        self.amount.pack(packer)
        packer.pack_uint(len(self.claimants))
        for claimants_item in self.claimants:
            claimants_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> CreateClaimableBalanceOp:
        asset = Asset.unpack(unpacker)
        amount = Int64.unpack(unpacker)
        length = unpacker.unpack_uint()
        claimants = []
        for _ in range(length):
            claimants.append(Claimant.unpack(unpacker))
        return cls(
            asset=asset,
            amount=amount,
            claimants=claimants,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> CreateClaimableBalanceOp:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> CreateClaimableBalanceOp:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.asset,
                self.amount,
                self.claimants,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.asset == other.asset
            and self.amount == other.amount
            and self.claimants == other.claimants
        )

    def __str__(self):
        out = [
            f"asset={self.asset}",
            f"amount={self.amount}",
            f"claimants={self.claimants}",
        ]
        return f"<CreateClaimableBalanceOp [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .bucket_metadata_ext import BucketMetadataExt
from .uint32 import Uint32

__all__ = ["BucketMetadata"]


class BucketMetadata:
    """
    XDR Source Code::

        struct BucketMetadata
        {
            // Indicates the protocol version used to create / merge this bucket.
            uint32 ledgerVersion;

            // reserved for future use
            union switch (int v)
            {
            case 0:
                void;
            }
            ext;
        };
    """

    def __init__(
        self,
        ledger_version: Uint32,
        ext: BucketMetadataExt,
    ) -> None:
        self.ledger_version = ledger_version
        self.ext = ext

    def pack(self, packer: Packer) -> None:
        self.ledger_version.pack(packer)
        self.ext.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> BucketMetadata:
        ledger_version = Uint32.unpack(unpacker)
        ext = BucketMetadataExt.unpack(unpacker)
        return cls(
            ledger_version=ledger_version,
            ext=ext,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> BucketMetadata:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> BucketMetadata:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ledger_version,
                self.ext,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.ledger_version == other.ledger_version and self.ext == other.ext

    def __str__(self):
        out = [
            f"ledger_version={self.ledger_version}",
            f"ext={self.ext}",
        ]
        return f"<BucketMetadata [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .soroban_authorized_invocation import SorobanAuthorizedInvocation
from .soroban_credentials import SorobanCredentials

__all__ = ["SorobanAuthorizationEntry"]


class SorobanAuthorizationEntry:
    """
    XDR Source Code::

        struct SorobanAuthorizationEntry
        {
            SorobanCredentials credentials;
            SorobanAuthorizedInvocation rootInvocation;
        };
    """

    def __init__(
        self,
        credentials: SorobanCredentials,
        root_invocation: SorobanAuthorizedInvocation,
    ) -> None:
        self.credentials = credentials
        self.root_invocation = root_invocation

    def pack(self, packer: Packer) -> None:
        self.credentials.pack(packer)
        self.root_invocation.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SorobanAuthorizationEntry:
        credentials = SorobanCredentials.unpack(unpacker)
        root_invocation = SorobanAuthorizedInvocation.unpack(unpacker)
        return cls(
            credentials=credentials,
            root_invocation=root_invocation,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SorobanAuthorizationEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SorobanAuthorizationEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.credentials,
                self.root_invocation,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.credentials == other.credentials
            and self.root_invocation == other.root_invocation
        )

    def __str__(self):
        out = [
            f"credentials={self.credentials}",
            f"root_invocation={self.root_invocation}",
        ]
        return f"<SorobanAuthorizationEntry [{', '.join(out)}]>"



from typing import Union

from ...call_builder.base.base_call_builder import BaseCallBuilder

__all__ = ["BasePaymentsCallBuilder"]


class BasePaymentsCallBuilder(BaseCallBuilder):
    """Creates a new :class:`PaymentsCallBuilder` pointed to server defined by horizon_url.

    See `List All Payments <https://developers.stellar.org/api/resources/operations/list-payments/>`__ for more information.

    :param horizon_url: Horizon server URL.
    """

    def __init__(self, horizon_url: str) -> None:
        super().__init__(horizon_url)
        self.endpoint: str = "payments"

    def for_account(self, account_id: str):
        """This endpoint responds with a collection of Payment operations where the given account
        was either the sender or receiver.

        See `Retrieve an Account's Payments <https://developers.stellar.org/api/resources/accounts/payments/>`__ for more information.

        :param account_id: Account ID
        :return: current PaymentsCallBuilder instance
        """
        self.endpoint = f"accounts/{account_id}/payments"
        return self

    def for_ledger(self, sequence: Union[int, str]):
        """This endpoint represents all payment operations that are part of a valid transactions in a given ledger.

        See `Retrieve a Ledger's Payments <https://developers.stellar.org/api/resources/ledgers/payments/>`__ for more information.

        :param sequence: Ledger sequence
        :return: current PaymentsCallBuilder instance
        """
        self.endpoint = f"ledgers/{sequence}/payments"
        return self

    def for_transaction(self, transaction_hash: str):
        """This endpoint represents all payment operations that are part of a given transaction.

        P.S. The documentation provided by SDF seems to be missing this API.

        :param transaction_hash: Transaction hash
        :return: current PaymentsCallBuilder instance
        """
        self.endpoint = f"transactions/{transaction_hash}/payments"
        return self

    def include_failed(self, include_failed: bool):
        """Adds a parameter defining whether to include failed transactions. By default only
        payments of successful transactions are returned.

        :param include_failed: Set to ``True`` to include payments of failed transactions.
        :return: current PaymentsCallBuilder instance
        """
        self._add_query_param("include_failed", include_failed)
        return self

    def join(self, join: str):
        """join represents `join` param in queries, currently only supports `transactions`

        :param join: join represents `join` param in queries, currently only supports `transactions`
        :return: current OperationsCallBuilder instance
        """
        self._add_query_param("join", join)
        return self



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .claim_atom_type import ClaimAtomType
from .claim_liquidity_atom import ClaimLiquidityAtom
from .claim_offer_atom import ClaimOfferAtom
from .claim_offer_atom_v0 import ClaimOfferAtomV0

__all__ = ["ClaimAtom"]


class ClaimAtom:
    """
    XDR Source Code::

        union ClaimAtom switch (ClaimAtomType type)
        {
        case CLAIM_ATOM_TYPE_V0:
            ClaimOfferAtomV0 v0;
        case CLAIM_ATOM_TYPE_ORDER_BOOK:
            ClaimOfferAtom orderBook;
        case CLAIM_ATOM_TYPE_LIQUIDITY_POOL:
            ClaimLiquidityAtom liquidityPool;
        };
    """

    def __init__(
        self,
        type: ClaimAtomType,
        v0: ClaimOfferAtomV0 = None,
        order_book: ClaimOfferAtom = None,
        liquidity_pool: ClaimLiquidityAtom = None,
    ) -> None:
        self.type = type
        self.v0 = v0
        self.order_book = order_book
        self.liquidity_pool = liquidity_pool

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == ClaimAtomType.CLAIM_ATOM_TYPE_V0:
            if self.v0 is None:
                raise ValueError("v0 should not be None.")
            self.v0.pack(packer)
            return
        if self.type == ClaimAtomType.CLAIM_ATOM_TYPE_ORDER_BOOK:
            if self.order_book is None:
                raise ValueError("order_book should not be None.")
            self.order_book.pack(packer)
            return
        if self.type == ClaimAtomType.CLAIM_ATOM_TYPE_LIQUIDITY_POOL:
            if self.liquidity_pool is None:
                raise ValueError("liquidity_pool should not be None.")
            self.liquidity_pool.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ClaimAtom:
        type = ClaimAtomType.unpack(unpacker)
        if type == ClaimAtomType.CLAIM_ATOM_TYPE_V0:
            v0 = ClaimOfferAtomV0.unpack(unpacker)
            return cls(type=type, v0=v0)
        if type == ClaimAtomType.CLAIM_ATOM_TYPE_ORDER_BOOK:
            order_book = ClaimOfferAtom.unpack(unpacker)
            return cls(type=type, order_book=order_book)
        if type == ClaimAtomType.CLAIM_ATOM_TYPE_LIQUIDITY_POOL:
            liquidity_pool = ClaimLiquidityAtom.unpack(unpacker)
            return cls(type=type, liquidity_pool=liquidity_pool)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ClaimAtom:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ClaimAtom:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.v0,
                self.order_book,
                self.liquidity_pool,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.v0 == other.v0
            and self.order_book == other.order_book
            and self.liquidity_pool == other.liquidity_pool
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(f"v0={self.v0}") if self.v0 is not None else None
        out.append(
            f"order_book={self.order_book}"
        ) if self.order_book is not None else None
        out.append(
            f"liquidity_pool={self.liquidity_pool}"
        ) if self.liquidity_pool is not None else None
        return f"<ClaimAtom [{', '.join(out)}]>"



import json
import logging
import time
from typing import Any, Dict, Generator, Optional, Tuple

import requests
from requests import RequestException, Session
from requests.adapters import DEFAULT_POOLSIZE, HTTPAdapter
from requests_sse import EventSource
from urllib3.exceptions import NewConnectionError
from urllib3.util import Retry

from ..__version__ import __version__
from ..client.base_sync_client import BaseSyncClient
from ..client.response import Response
from ..exceptions import ConnectionError, StreamClientError
from . import defines

DEFAULT_NUM_RETRIES = 3
DEFAULT_BACKOFF_FACTOR = 0.5
USER_AGENT = f"py-stellar-base/{__version__}/RequestsClient"
IDENTIFICATION_HEADERS = {
    "X-Client-Name": "py-stellar-base",
    "X-Client-Version": __version__,
}

logger = logging.getLogger(__name__)

__all__ = ["RequestsClient"]


class RequestsClient(BaseSyncClient):
    """The :class:`RequestsClient` object is a synchronous http client,
    which represents the interface for making requests to a server instance.

    :param pool_size: persistent connection to Horizon and connection pool
    :param num_retries: configurable request retry functionality
    :param request_timeout: the timeout for all GET requests (for each retry)
    :param post_timeout: the timeout for all POST requests (for each retry)
    :param backoff_factor: a backoff factor to apply between attempts after the second try
    :param session: the request session
    :param stream_session: the stream request session
    :param custom_headers: any additional HTTP headers to add in requests
    """

    def __init__(
        self,
        pool_size: int = DEFAULT_POOLSIZE,
        num_retries: int = DEFAULT_NUM_RETRIES,
        request_timeout: int = defines.DEFAULT_GET_TIMEOUT_SECONDS,
        post_timeout: float = defines.DEFAULT_POST_TIMEOUT_SECONDS,
        backoff_factor: float = DEFAULT_BACKOFF_FACTOR,
        session: Session = None,
        stream_session: Session = None,
        custom_headers: Optional[Dict[str, str]] = None,
    ):
        self.pool_size: int = pool_size
        self.num_retries: int = num_retries
        self.request_timeout: int = request_timeout
        self.post_timeout: float = post_timeout
        self.backoff_factor: float = backoff_factor

        # adding 504 to the tuple of statuses to retry
        self.status_forcelist: Tuple[int] = tuple(Retry.RETRY_AFTER_STATUS_CODES) + (
            504,
        )  # type: ignore[assignment]

        # configure standard session

        # configure retry handler
        retry = Retry(
            total=self.num_retries,
            backoff_factor=self.backoff_factor,
            redirect=0,
            status_forcelist=self.status_forcelist,
            allowed_methods=frozenset(["GET", "POST"]),
            raise_on_status=False,
        )
        # init transport adapter
        adapter = HTTPAdapter(
            pool_connections=self.pool_size,
            pool_maxsize=self.pool_size,
            max_retries=retry,
        )

        self.headers = {**IDENTIFICATION_HEADERS, "User-Agent": USER_AGENT}

        if custom_headers:
            self.headers = {**self.headers, **custom_headers}

        # init session
        if session is None:
            session = requests.Session()

            # set default headers
            session.headers.update(self.headers)

            session.mount("http://", adapter)
            session.mount("https://", adapter)
        self._session: Session = session
        self._stream_session: Optional[Session] = stream_session

    def get(self, url: str, params: Dict[str, str] = None) -> Response:
        """Perform HTTP GET request.

        :param url: the request url
        :param params: the request params
        :return: the response from server
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        try:
            resp = self._session.get(url, params=params, timeout=self.request_timeout)
        except (RequestException, NewConnectionError) as err:
            raise ConnectionError(err)
        return Response(
            status_code=resp.status_code,
            text=resp.text,
            headers=dict(resp.headers),
            url=resp.url,
        )

    def post(
        self, url: str, data: Dict[str, str] = None, json_data: Dict[str, Any] = None
    ) -> Response:
        """Perform HTTP POST request.

        :param url: the request url
        :param data: the data send to server
        :param json_data: the json data send to server
        :return: the response from server
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        try:
            resp = self._session.post(
                url, data=data, json=json_data, timeout=self.post_timeout
            )
        except (RequestException, NewConnectionError) as err:
            raise ConnectionError(err)
        return Response(
            status_code=resp.status_code,
            text=resp.text,
            headers=dict(resp.headers),
            url=resp.url,
        )

    def stream(
        self, url: str, params: Dict[str, str] = None
    ) -> Generator[Dict[str, Any], None, None]:
        """Creates an EventSource that listens for incoming messages from the server.

        See `Horizon Response Format <https://developers.stellar.org/api/introduction/response-format/>`__

        See `MDN EventSource <https://developer.mozilla.org/en-US/docs/Web/API/EventSource>`_

        :param url: the request url
        :param params: the request params
        :return: a Generator for server response
        :raise: :exc:`StreamClientError <stellar_sdk.exceptions.StreamClientError>`
        """
        query_params = {**params} if params else dict()

        query_params.update(**IDENTIFICATION_HEADERS)
        retry = 0.1

        while True:
            try:
                """
                Create a new SSEClient:
                Using the last id as the cursor
                """
                with EventSource(
                    url,
                    timeout=60,
                    params=query_params,
                    headers=self.headers,
                ) as client:
                    for event in client:
                        if event.last_event_id:
                            query_params["cursor"] = event.last_event_id
                            # Events that dont have an id are not useful for us (hello/byebye events)
                        retry = client._reconnection_time.total_seconds()
                        try:
                            data = event.data
                            if data != '"hello"' and data != '"byebye"':
                                yield json.loads(data)
                        except json.JSONDecodeError:
                            # Content was not json-decodable
                            pass  # pragma: no cover
            except requests.Timeout:
                logger.warning(
                    f"We have encountered an timeout error and we will try to reconnect, cursor = {query_params.get('cursor')}"
                )
                time.sleep(retry)
            except requests.RequestException as e:
                raise StreamClientError(
                    query_params["cursor"], "Failed to get stream message."
                ) from e

    def close(self) -> None:
        """Close underlying connector.

        Release all acquired resources.
        """
        self._session.close()
        if self._stream_session:
            self._stream_session.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __str__(self):
        return (
            f"<RequestsClient [pool_size={self.pool_size}, "
            f"num_retries={self.num_retries}, "
            f"request_timeout={self.request_timeout}, "
            f"post_timeout={self.post_timeout}, "
            f"backoff_factor={self.backoff_factor}, "
            f"session={self._session}, "
            f"stream_session={self.backoff_factor}]>"
        )



class Ed25519PublicKeySigner:
    def __init__(self, account_id: str, weight: int = 0) -> None:
        """The :class:`Signer` object, which represents the signer for the client account.

        :param account_id: Account ID (ex. ``"GBYNR2QJXLBCBTRN44MRORCMI4YO7FZPFBCNOKTOBCAAFC7KC3LNPRYS"``)
        :param weight: The signer's weight.
        """
        self.account_id = account_id
        self.weight = weight

    def __hash__(self):
        return hash((self.account_id, self.weight))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.account_id == other.account_id and self.weight == other.weight

    def __str__(self):
        return f"<Ed25519PublicKeySigner [account_id={self.account_id}, weight={self.weight}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .base import String
from .constants import *
from .sc_spec_udt_enum_case_v0 import SCSpecUDTEnumCaseV0

__all__ = ["SCSpecUDTEnumV0"]


class SCSpecUDTEnumV0:
    """
    XDR Source Code::

        struct SCSpecUDTEnumV0
        {
            string doc<SC_SPEC_DOC_LIMIT>;
            string lib<80>;
            string name<60>;
            SCSpecUDTEnumCaseV0 cases<50>;
        };
    """

    def __init__(
        self,
        doc: bytes,
        lib: bytes,
        name: bytes,
        cases: List[SCSpecUDTEnumCaseV0],
    ) -> None:
        _expect_max_length = 50
        if cases and len(cases) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `cases` should be {_expect_max_length}, but got {len(cases)}."
            )
        self.doc = doc
        self.lib = lib
        self.name = name
        self.cases = cases

    def pack(self, packer: Packer) -> None:
        String(self.doc, SC_SPEC_DOC_LIMIT).pack(packer)
        String(self.lib, 80).pack(packer)
        String(self.name, 60).pack(packer)
        packer.pack_uint(len(self.cases))
        for cases_item in self.cases:
            cases_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCSpecUDTEnumV0:
        doc = String.unpack(unpacker)
        lib = String.unpack(unpacker)
        name = String.unpack(unpacker)
        length = unpacker.unpack_uint()
        cases = []
        for _ in range(length):
            cases.append(SCSpecUDTEnumCaseV0.unpack(unpacker))
        return cls(
            doc=doc,
            lib=lib,
            name=name,
            cases=cases,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCSpecUDTEnumV0:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCSpecUDTEnumV0:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.doc,
                self.lib,
                self.name,
                self.cases,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.doc == other.doc
            and self.lib == other.lib
            and self.name == other.name
            and self.cases == other.cases
        )

    def __str__(self):
        out = [
            f"doc={self.doc}",
            f"lib={self.lib}",
            f"name={self.name}",
            f"cases={self.cases}",
        ]
        return f"<SCSpecUDTEnumV0 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .asset_code4 import AssetCode4
from .asset_code12 import AssetCode12
from .asset_type import AssetType

__all__ = ["AssetCode"]


class AssetCode:
    """
    XDR Source Code::

        union AssetCode switch (AssetType type)
        {
        case ASSET_TYPE_CREDIT_ALPHANUM4:
            AssetCode4 assetCode4;

        case ASSET_TYPE_CREDIT_ALPHANUM12:
            AssetCode12 assetCode12;

            // add other asset types here in the future
        };
    """

    def __init__(
        self,
        type: AssetType,
        asset_code4: AssetCode4 = None,
        asset_code12: AssetCode12 = None,
    ) -> None:
        self.type = type
        self.asset_code4 = asset_code4
        self.asset_code12 = asset_code12

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == AssetType.ASSET_TYPE_CREDIT_ALPHANUM4:
            if self.asset_code4 is None:
                raise ValueError("asset_code4 should not be None.")
            self.asset_code4.pack(packer)
            return
        if self.type == AssetType.ASSET_TYPE_CREDIT_ALPHANUM12:
            if self.asset_code12 is None:
                raise ValueError("asset_code12 should not be None.")
            self.asset_code12.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> AssetCode:
        type = AssetType.unpack(unpacker)
        if type == AssetType.ASSET_TYPE_CREDIT_ALPHANUM4:
            asset_code4 = AssetCode4.unpack(unpacker)
            return cls(type=type, asset_code4=asset_code4)
        if type == AssetType.ASSET_TYPE_CREDIT_ALPHANUM12:
            asset_code12 = AssetCode12.unpack(unpacker)
            return cls(type=type, asset_code12=asset_code12)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> AssetCode:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> AssetCode:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.asset_code4,
                self.asset_code12,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.asset_code4 == other.asset_code4
            and self.asset_code12 == other.asset_code12
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(
            f"asset_code4={self.asset_code4}"
        ) if self.asset_code4 is not None else None
        out.append(
            f"asset_code12={self.asset_code12}"
        ) if self.asset_code12 is not None else None
        return f"<AssetCode [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .account_id import AccountID
from .claim_predicate import ClaimPredicate

__all__ = ["ClaimantV0"]


class ClaimantV0:
    """
    XDR Source Code::

        struct
            {
                AccountID destination;    // The account that can use this condition
                ClaimPredicate predicate; // Claimable if predicate is true
            }
    """

    def __init__(
        self,
        destination: AccountID,
        predicate: ClaimPredicate,
    ) -> None:
        self.destination = destination
        self.predicate = predicate

    def pack(self, packer: Packer) -> None:
        self.destination.pack(packer)
        self.predicate.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ClaimantV0:
        destination = AccountID.unpack(unpacker)
        predicate = ClaimPredicate.unpack(unpacker)
        return cls(
            destination=destination,
            predicate=predicate,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ClaimantV0:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ClaimantV0:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.destination,
                self.predicate,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.destination == other.destination and self.predicate == other.predicate
        )

    def __str__(self):
        out = [
            f"destination={self.destination}",
            f"predicate={self.predicate}",
        ]
        return f"<ClaimantV0 [{', '.join(out)}]>"



from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..muxed_account import MuxedAccount
from .operation import Operation

__all__ = ["BumpSequence"]


class BumpSequence(Operation):
    """The :class:`BumpSequence` object, which represents a
    BumpSequence operation on Stellar's network.

    Bump sequence allows to bump forward the sequence number of the source account of the
    operation, allowing to invalidate any transactions with a smaller sequence number.
    If the specified bumpTo sequence number is greater than the source account’s sequence number,
    the account’s sequence number is updated with that value, otherwise it’s not modified.

    Threshold: Low

    See `Bump Sequence <https://developers.stellar.org/docs/start/list-of-operations/#bump-sequence>`_ for more information.

    :param bump_to: Sequence number to bump to.
    :param source: The optional source account.

    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.BUMP_SEQUENCE
    )

    def __init__(
        self, bump_to: int, source: Optional[Union[MuxedAccount, str]] = None
    ) -> None:
        super().__init__(source)
        self.bump_to: int = bump_to

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        sequence = stellar_xdr.SequenceNumber(stellar_xdr.Int64(self.bump_to))
        bump_sequence_op = stellar_xdr.BumpSequenceOp(sequence)
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE, bump_sequence_op=bump_sequence_op
        )
        return body

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Operation) -> "BumpSequence":
        """Creates a :class:`BumpSequence` object from an XDR Operation
        object.

        """
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.bump_sequence_op is not None
        bump_to = xdr_object.body.bump_sequence_op.bump_to.sequence_number.int64
        op = cls(source=source, bump_to=bump_to)
        return op

    def __str__(self):
        return f"<BumpSequence [bump_to={self.bump_to}, source={self.source}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .hash import Hash
from .uint32 import Uint32

__all__ = ["TTLEntry"]


class TTLEntry:
    """
    XDR Source Code::

        struct TTLEntry {
            // Hash of the LedgerKey that is associated with this TTLEntry
            Hash keyHash;
            uint32 liveUntilLedgerSeq;
        };
    """

    def __init__(
        self,
        key_hash: Hash,
        live_until_ledger_seq: Uint32,
    ) -> None:
        self.key_hash = key_hash
        self.live_until_ledger_seq = live_until_ledger_seq

    def pack(self, packer: Packer) -> None:
        self.key_hash.pack(packer)
        self.live_until_ledger_seq.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TTLEntry:
        key_hash = Hash.unpack(unpacker)
        live_until_ledger_seq = Uint32.unpack(unpacker)
        return cls(
            key_hash=key_hash,
            live_until_ledger_seq=live_until_ledger_seq,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TTLEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TTLEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.key_hash,
                self.live_until_ledger_seq,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.key_hash == other.key_hash
            and self.live_until_ledger_seq == other.live_until_ledger_seq
        )

    def __str__(self):
        out = [
            f"key_hash={self.key_hash}",
            f"live_until_ledger_seq={self.live_until_ledger_seq}",
        ]
        return f"<TTLEntry [{', '.join(out)}]>"



from ...asset import Asset
from ...call_builder.base import BaseOrderbookCallBuilder
from ...call_builder.call_builder_sync.base_call_builder import BaseCallBuilder
from ...client.base_sync_client import BaseSyncClient

__all__ = ["OrderbookCallBuilder"]


class OrderbookCallBuilder(BaseCallBuilder, BaseOrderbookCallBuilder):
    """Creates a new :class:`OrderbookCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.Server.orderbook`.

    See `Orderbook <https://developers.stellar.org/api/aggregations/order-books/>`__ for more information.

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    :param selling: Asset being sold
    :param buying: Asset being bought
    """

    def __init__(
        self,
        horizon_url: str,
        client: BaseSyncClient,
        selling: Asset,
        buying: Asset,
    ) -> None:
        super().__init__(
            horizon_url=horizon_url, client=client, selling=selling, buying=buying
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .message_type import MessageType
from .uint256 import Uint256

__all__ = ["DontHave"]


class DontHave:
    """
    XDR Source Code::

        struct DontHave
        {
            MessageType type;
            uint256 reqHash;
        };
    """

    def __init__(
        self,
        type: MessageType,
        req_hash: Uint256,
    ) -> None:
        self.type = type
        self.req_hash = req_hash

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        self.req_hash.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> DontHave:
        type = MessageType.unpack(unpacker)
        req_hash = Uint256.unpack(unpacker)
        return cls(
            type=type,
            req_hash=req_hash,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> DontHave:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> DontHave:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.req_hash,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.type == other.type and self.req_hash == other.req_hash

    def __str__(self):
        out = [
            f"type={self.type}",
            f"req_hash={self.req_hash}",
        ]
        return f"<DontHave [{', '.join(out)}]>"



"""
SEP: 0010
Title: Stellar Web Authentication
Author: Sergey Nebolsin <@nebolsin>, Tom Quisel <tom.quisel@gmail.com>, Leigh McCulloch <@leighmcculloch>, Jake Urban <jake@stellar.org>
Status: Active
Created: 2018-07-31
Updated: 2021-08-10
Version 3.3.0
"""
import base64
import os
import time
from typing import Iterable, List, Optional, Sequence, Union

from .. import xdr as stellar_xdr
from ..account import Account
from ..exceptions import BadSignatureError
from ..keypair import Keypair
from ..memo import IdMemo, NoneMemo
from ..muxed_account import MuxedAccount
from ..operation.manage_data import ManageData
from ..transaction_builder import TransactionBuilder
from ..transaction_envelope import TransactionEnvelope
from .ed25519_public_key_signer import Ed25519PublicKeySigner
from .exceptions import InvalidSep10ChallengeError

__all__ = [
    "build_challenge_transaction",
    "verify_challenge_transaction_signers",
    "verify_challenge_transaction_signed_by_client_master_key",
    "verify_challenge_transaction_threshold",
    "read_challenge_transaction",
    "verify_challenge_transaction",
]

MUXED_ACCOUNT_STARTING_LETTER: str = "M"


class ChallengeTransaction:
    """Used to store the results produced
    by :func:`stellar_sdk.sep.stellar_web_authentication.read_challenge_transaction`.

    :param transaction: The TransactionEnvelope parsed from challenge xdr.
    :param client_account_id: The stellar account that the wallet wishes to authenticate with the server.
    :param matched_home_domain: The domain name that has been matched.
    :param memo: The ID memo attached to the transaction
    """

    def __init__(
        self,
        transaction: TransactionEnvelope,
        client_account_id: str,
        matched_home_domain: str,
        memo: Optional[int] = None,
    ) -> None:
        self.transaction = transaction
        self.client_account_id = client_account_id
        self.matched_home_domain = matched_home_domain
        self.memo = memo

    def __hash__(self):
        return hash(
            (
                self.transaction,
                self.client_account_id,
                self.matched_home_domain,
                self.memo,
            )
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.transaction == other.transaction
            and self.client_account_id == other.client_account_id
            and self.matched_home_domain == other.matched_home_domain
            and self.memo == other.memo
        )

    def __str__(self):
        return f"<ChallengeTransaction [transaction={self.transaction}, client_account_id={self.client_account_id}, memo={self.memo}, matched_home_domain={self.matched_home_domain}]>"


def build_challenge_transaction(
    server_secret: str,
    client_account_id: str,
    home_domain: str,
    web_auth_domain: str,
    network_passphrase: str,
    timeout: int = 900,
    client_domain: Optional[str] = None,
    client_signing_key: Optional[str] = None,
    memo: Optional[int] = None,
) -> str:
    """Returns a valid `SEP0010 <https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0010.md>`_
    challenge transaction which you can use for Stellar Web Authentication.

    :param server_secret: secret key for server's stellar.toml `SIGNING_KEY`.
    :param client_account_id: The stellar account (``G...``) or
        muxed account (``M...``) that the wallet wishes to authenticate with the server.
    :param home_domain: The `fully qualified domain name <https://en.wikipedia.org/wiki/Fully_qualified_home_domain>`_
        of the service requiring authentication (ex. ``"example.com"``).
    :param web_auth_domain: The fully qualified domain name of the service issuing the challenge.
    :param network_passphrase: The network to connect to for verifying and retrieving
        additional attributes from. (ex. ``"Public Global Stellar Network ; September 2015"``)
    :param timeout: Challenge duration in seconds (default to 15 minutes).
    :param client_domain: The domain of the client application requesting authentication
    :param client_signing_key: The stellar account listed as the SIGNING_KEY on the client domain's TOML file
    :param memo: The ID memo to attach to the transaction. Not permitted if `client_account_id` is a muxed account
    :return: A base64 encoded string of the raw TransactionEnvelope xdr struct for the transaction.
    """
    if client_account_id.startswith(MUXED_ACCOUNT_STARTING_LETTER) and memo:
        raise ValueError(
            "memos are not valid for challenge transactions with a muxed client account"
        )

    now = int(time.time())
    server_keypair = Keypair.from_secret(server_secret)
    server_account = Account(account=server_keypair.public_key, sequence=-1)
    transaction_builder = TransactionBuilder(server_account, network_passphrase, 100)
    transaction_builder.add_time_bounds(min_time=now, max_time=now + timeout)
    nonce = os.urandom(48)
    nonce_encoded = base64.b64encode(nonce)
    transaction_builder.append_manage_data_op(
        data_name=f"{home_domain} auth",
        data_value=nonce_encoded,
        source=client_account_id,
    ).append_manage_data_op(
        data_name="web_auth_domain",
        data_value=web_auth_domain,
        source=server_account.account,
    )
    if client_domain:
        if not client_signing_key:
            raise ValueError(
                "client_signing_key is required if client_domain is provided."
            )
        transaction_builder.append_manage_data_op(
            data_name="client_domain",
            data_value=client_domain,
            source=client_signing_key,
        )
    if memo:
        transaction_builder.add_id_memo(memo)
    transaction = transaction_builder.build()
    transaction.sign(server_keypair)
    return transaction.to_xdr()


def read_challenge_transaction(
    challenge_transaction: str,
    server_account_id: str,
    home_domains: Union[str, Iterable[str]],
    web_auth_domain: str,
    network_passphrase: str,
) -> ChallengeTransaction:
    """Reads a SEP 10 challenge transaction and returns the decoded transaction envelope and client account ID contained within.

    It also verifies that transaction is signed by the server.

    It does not verify that the transaction has been signed by the client or
    that any signatures other than the servers on the transaction are valid. Use
    one of the following functions to completely verify the transaction:

    * :func:`stellar_sdk.sep.stellar_web_authentication.verify_challenge_transaction_threshold`
    * :func:`stellar_sdk.sep.stellar_web_authentication.verify_challenge_transaction_signers`

    :param challenge_transaction: SEP0010 transaction challenge transaction in base64.
    :param server_account_id: public key for server's account.
    :param home_domains: The home domain that is expected to be included in the first Manage Data operation's string
        key. If a list is provided, one of the domain names in the array must match.
    :param web_auth_domain: The home domain that is expected to be included as the value of the Manage Data operation with
        the 'web_auth_domain' key. If no such operation is included, this parameter is not used.
    :param network_passphrase: The network to connect to for verifying and retrieving
        additional attributes from. (ex. ``"Public Global Stellar Network ; September 2015"``)
    :raises: :exc:`InvalidSep10ChallengeError <stellar_sdk.sep.exceptions.InvalidSep10ChallengeError>` - if the
        validation fails, the exception will be thrown.
    """

    # decode the received input as a base64-urlencoded XDR representation of Stellar transaction envelope
    if server_account_id.startswith(MUXED_ACCOUNT_STARTING_LETTER):
        raise ValueError(
            "Invalid server_account_id, multiplexed account are not supported."
        )

    xdr_object = stellar_xdr.TransactionEnvelope.from_xdr(challenge_transaction)
    if xdr_object.type == stellar_xdr.EnvelopeType.ENVELOPE_TYPE_TX_FEE_BUMP:
        raise ValueError(
            "Invalid challenge, expected a TransactionEnvelope but received a FeeBumpTransactionEnvelope."
        )

    try:
        transaction_envelope = TransactionEnvelope.from_xdr(
            challenge_transaction, network_passphrase=network_passphrase
        )
    except Exception:
        raise InvalidSep10ChallengeError(
            "Importing XDR failed, please check if XDR is correct."
        )

    transaction = transaction_envelope.transaction

    # verify that transaction source account is equal to the server's signing key
    if transaction.source.account_id != server_account_id:
        raise InvalidSep10ChallengeError(
            "Transaction source account is not equal to server's account."
        )

    # verify that transaction sequenceNumber is equal to zero
    if transaction.sequence != 0:
        raise InvalidSep10ChallengeError(
            "The transaction sequence number should be zero."
        )

    # verify that transaction has time bounds set, and that current time is between the minimum and maximum bounds
    if not transaction.preconditions or not transaction.preconditions.time_bounds:
        raise InvalidSep10ChallengeError("Transaction requires timebounds.")

    max_time = transaction.preconditions.time_bounds.max_time
    min_time = transaction.preconditions.time_bounds.min_time

    if max_time == 0:
        raise InvalidSep10ChallengeError(
            "Transaction requires non-infinite timebounds."
        )

    current_time = time.time()
    # Apply a grace period to the challenge MinTime to account for
    # clock drift between the server and client
    # https://github.com/StellarCN/py-stellar-base/issues/524
    grace_period = 60 * 5
    if current_time < min_time - grace_period or current_time > max_time:
        raise InvalidSep10ChallengeError(
            "Transaction is not within range of the specified timebounds."
        )

    # verify that transaction contains a single Manage Data operation and its source account is not null
    if len(transaction.operations) < 1:
        raise InvalidSep10ChallengeError(
            "Transaction should contain at least one operation."
        )

    manage_data_op = transaction.operations[0]
    if not isinstance(manage_data_op, ManageData):
        raise InvalidSep10ChallengeError("Operation type should be ManageData.")

    client_account = manage_data_op.source
    if not client_account:
        raise InvalidSep10ChallengeError("Operation should have a source account.")

    client_account_address = client_account.account_muxed or client_account.account_id
    matched_home_domain = None
    if isinstance(home_domains, str):
        if manage_data_op.data_name == f"{home_domains} auth":
            matched_home_domain = home_domains
    else:
        for home_domain in home_domains:
            if manage_data_op.data_name == f"{home_domain} auth":
                matched_home_domain = home_domain
                break

    if matched_home_domain is None:
        raise InvalidSep10ChallengeError(
            "The transaction's operation key name does not "
            "include the expected home domain."
        )

    if manage_data_op.data_value is None:
        raise InvalidSep10ChallengeError("Operation value should not be null.")

    if len(manage_data_op.data_value) != 64:
        raise InvalidSep10ChallengeError(
            "Operation value encoded as base64 should be 64 bytes long."
        )

    nonce = base64.b64decode(manage_data_op.data_value)
    if len(nonce) != 48:
        raise InvalidSep10ChallengeError(
            "Operation value before encoding as base64 should be 48 bytes long."
        )

    if not transaction.memo or isinstance(transaction.memo, NoneMemo):
        memo = None
    elif client_account.account_muxed_id:
        raise InvalidSep10ChallengeError(
            "Invalid challenge, memos are not permitted if the client account is muxed"
        )
    elif isinstance(transaction.memo, IdMemo):
        memo = transaction.memo.memo_id
    else:
        raise InvalidSep10ChallengeError("Invalid memo, only ID memos are permitted")

    # verify any subsequent operations are manage data ops and source account is the server
    for op in transaction.operations[1:]:
        if not isinstance(op, ManageData):
            raise InvalidSep10ChallengeError("Operation type should be ManageData.")
        if op.source is None:
            raise InvalidSep10ChallengeError("Operation should have a source account.")
        if (
            op.source.account_id != server_account_id
            and op.data_name != "client_domain"
        ):
            raise InvalidSep10ChallengeError(
                "The transaction has operations that are unrecognized."
            )
        if op.data_name == "web_auth_domain":
            if op.data_value is None:
                raise InvalidSep10ChallengeError(
                    "'web_auth_domain' operation value should not be null."
                )
            if op.data_value != web_auth_domain.encode():
                raise InvalidSep10ChallengeError(
                    f"'web_auth_domain' operation value does not match {web_auth_domain}."
                )

    # verify that transaction envelope has a correct signature by server's signing key
    if not _verify_te_signed_by_account_id(transaction_envelope, server_account_id):
        raise InvalidSep10ChallengeError(
            f"Transaction not signed by server: {server_account_id}."
        )

    return ChallengeTransaction(
        transaction=transaction_envelope,
        client_account_id=client_account_address,
        matched_home_domain=matched_home_domain,
        memo=memo,
    )


def verify_challenge_transaction_signers(
    challenge_transaction: str,
    server_account_id: str,
    home_domains: Union[str, Iterable[str]],
    web_auth_domain: str,
    network_passphrase: str,
    signers: Sequence[Ed25519PublicKeySigner],
) -> List[Ed25519PublicKeySigner]:
    """Verifies that for a SEP 10 challenge transaction
    all signatures on the transaction are accounted for. A transaction is
    verified if it is signed by the server account, and all other signatures
    match a signer that has been provided as an argument. Additional signers can
    be provided that do not have a signature, but all signatures must be matched
    to a signer for verification to succeed. If verification succeeds a list of
    signers that were found is returned, excluding the server account ID.

    :param challenge_transaction: SEP0010 transaction challenge transaction in base64.
    :param server_account_id: public key for server's account.
    :param home_domains: The home domain that is expected to be included in the first Manage Data operation's string
        key. If a list is provided, one of the domain names in the array must match.
    :param web_auth_domain: The home domain that is expected to be included as the value of the Manage Data
        operation with the 'web_auth_domain' key, if present.
    :param network_passphrase: The network to connect to for verifying and retrieving
        additional attributes from. (ex. ``"Public Global Stellar Network ; September 2015"``)
    :param signers: The signers of client account.

    :raises: :exc:`InvalidSep10ChallengeError <stellar_sdk.sep.exceptions.InvalidSep10ChallengeError>`:
        - The transaction is invalid according to :func:`stellar_sdk.sep.stellar_web_authentication.read_challenge_transaction`.
        - One or more signatures in the transaction are not identifiable as the server account or one of the signers provided in the arguments.
    """
    if not signers:
        raise InvalidSep10ChallengeError("No signers provided.")

    parsed_challenge_transaction = read_challenge_transaction(
        challenge_transaction,
        server_account_id,
        home_domains,
        web_auth_domain,
        network_passphrase,
    )
    te = parsed_challenge_transaction.transaction
    server_keypair = Keypair.from_public_key(server_account_id)

    # If the client domain is included the challenge transaction,
    # verify that the transaction is signed by the operation's source account.
    client_signing_key = None
    for operation in te.transaction.operations:
        if isinstance(operation, ManageData) and operation.data_name == "client_domain":
            client_signing_key = operation.source
            break

    # Ensure the server is not included
    # anywhere we check or output the list of signers.
    client_signers = []
    for signer in signers:
        # Ignore the server signer if it is in the signers list. It's
        # important when verifying signers of a challenge transaction that we
        # only verify and return client signers. If an account has the server
        # as a signer the server should not play a part in the authentication
        # of the client.
        if signer == server_keypair.public_key:
            continue
        client_signers.append(signer)

    # Verify all the transaction's signers (server and client) in one
    # hit. We do this in one hit here even though the server signature was
    # checked in the read_challenge_transaction to ensure that every signature and signer
    # are consumed only once on the transaction.
    additional_signers = [Ed25519PublicKeySigner(server_keypair.public_key)]
    if client_signing_key:
        additional_signers.append(Ed25519PublicKeySigner(client_signing_key.account_id))
    all_signers = client_signers + additional_signers
    all_signers_found = _verify_transaction_signatures(te, all_signers)

    signers_found: List[Ed25519PublicKeySigner] = []
    server_signer_found = False
    client_signing_key_found = False
    for signer in all_signers_found:
        if signer.account_id == server_keypair.public_key:
            server_signer_found = True
            continue
        if client_signing_key and signer.account_id == client_signing_key.account_id:
            client_signing_key_found = True
            continue
        # Deduplicate the client signers
        if _signer_in_signers(signer, signers_found):
            continue
        signers_found.append(signer)

    # Confirm we matched a signature to the server signer.
    if not server_signer_found:
        raise InvalidSep10ChallengeError(
            f"Transaction not signed by server: {server_keypair.public_key}."
        )

    if client_signing_key and not client_signing_key_found:
        raise InvalidSep10ChallengeError(
            "Transaction not signed by the source account of the 'client_domain' "
            "ManageData operation"
        )

    # Confirm we matched signatures to the client signers.
    if not signers_found:
        raise InvalidSep10ChallengeError("Transaction not signed by any client signer.")

    # Confirm all signatures were consumed by a signer.
    if len(all_signers_found) != len(te.signatures):
        raise InvalidSep10ChallengeError("Transaction has unrecognized signatures.")

    return signers_found


def verify_challenge_transaction_signed_by_client_master_key(
    challenge_transaction: str,
    server_account_id: str,
    home_domains: Union[str, Iterable[str]],
    web_auth_domain: str,
    network_passphrase: str,
) -> None:
    """An alias for :func:`stellar_sdk.sep.stellar_web_authentication.verify_challenge_transaction`.

    :param challenge_transaction: SEP0010 transaction challenge transaction in base64.
    :param server_account_id: public key for server's account.
    :param home_domains: The home domain that is expected to be included in the first Manage Data operation's string
        key. If a list is provided, one of the domain names in the array must match.
    :param web_auth_domain: The home domain that is expected to be included as the value of the Manage Data operation with
        the 'web_auth_domain' key. If no such operation is included, this parameter is not used.
    :param network_passphrase: The network to connect to for verifying and retrieving
        additional attributes from. (ex. ``"Public Global Stellar Network ; September 2015"``)

    :raises: :exc:`InvalidSep10ChallengeError <stellar_sdk.sep.exceptions.InvalidSep10ChallengeError>` - if the
        validation fails, the exception will be thrown.
    """

    return verify_challenge_transaction(
        challenge_transaction,
        server_account_id,
        home_domains,
        web_auth_domain,
        network_passphrase,
    )


def verify_challenge_transaction_threshold(
    challenge_transaction: str,
    server_account_id: str,
    home_domains: Union[str, Iterable[str]],
    web_auth_domain: str,
    network_passphrase: str,
    threshold: int,
    signers: Sequence[Ed25519PublicKeySigner],
) -> List[Ed25519PublicKeySigner]:
    """Verifies that for a SEP 10 challenge transaction
    all signatures on the transaction are accounted for and that the signatures
    meet a threshold on an account. A transaction is verified if it is signed by
    the server account, and all other signatures match a signer that has been
    provided as an argument, and those signatures meet a threshold on the
    account.

    :param challenge_transaction: SEP0010 transaction challenge transaction in base64.
    :param server_account_id: public key for server's account.
    :param home_domains: The home domain that is expected to be included in the first Manage Data operation's string
        key. If a list is provided, one of the domain names in the array must match.
    :param web_auth_domain: The home domain that is expected to be included as the value of the Manage Data operation with
        the 'web_auth_domain' key. If no such operation is included, this parameter is not used.
    :param network_passphrase: The network to connect to for verifying and retrieving
        additional attributes from. (ex. ``"Public Global Stellar Network ; September 2015"``)
    :param threshold: The medThreshold on the client account.
    :param signers: The signers of client account.

    :raises: :exc:`InvalidSep10ChallengeError <stellar_sdk.sep.exceptions.InvalidSep10ChallengeError>`:
        - The transaction is invalid according to :func:`stellar_sdk.sep.stellar_web_authentication.read_challenge_transaction`.
        - One or more signatures in the transaction are not identifiable as the server account or one of the signers provided in the arguments.
        - The signatures are all valid but do not meet the threshold.
    """
    signers_found = verify_challenge_transaction_signers(
        challenge_transaction,
        server_account_id,
        home_domains,
        web_auth_domain,
        network_passphrase,
        signers,
    )

    weight = sum(signer.weight for signer in signers_found)
    if weight < threshold:
        raise InvalidSep10ChallengeError(
            f"signers with weight {weight} do not meet threshold {threshold}."
        )

    return signers_found


def verify_challenge_transaction(
    challenge_transaction: str,
    server_account_id: str,
    home_domains: Union[str, Iterable[str]],
    web_auth_domain: str,
    network_passphrase: str,
) -> None:
    """Verifies if a transaction is a valid
    `SEP0010 v1.2 <https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0010.md>`_
    challenge transaction, if the validation fails, an exception will be thrown.

    This function performs the following checks:

        1. verify that transaction sequenceNumber is equal to zero;
        2. verify that transaction source account is equal to the server's signing key;
        3. verify that transaction has time bounds set, and that current time is between the minimum and maximum bounds;
        4. verify that transaction contains a single Manage Data operation and it's source account is not null;
        5. verify that transaction envelope has a correct signature by server's signing key;
        6. verify that transaction envelope has a correct signature by the operation's source account;

    :param challenge_transaction: SEP0010 transaction challenge transaction in base64.
    :param server_account_id: public key for server's account.
    :param home_domains: The home domain that is expected to be included in the first Manage Data operation's string
        key. If a list is provided, one of the domain names in the array must match.
    :param web_auth_domain: The home domain that is expected to be included as the value of the Manage Data
        operation with the `web_auth_domain` key, if present.
    :param network_passphrase: The network to connect to for verifying and retrieving
        additional attributes from. (ex. ``"Public Global Stellar Network ; September 2015"``)
    :raises: :exc:`InvalidSep10ChallengeError <stellar_sdk.sep.exceptions.InvalidSep10ChallengeError>` - if the
        validation fails, the exception will be thrown.
    """

    parsed_challenge_transaction = read_challenge_transaction(
        challenge_transaction,
        server_account_id,
        home_domains,
        web_auth_domain,
        network_passphrase,
    )
    client_account_id = parsed_challenge_transaction.client_account_id
    if client_account_id.startswith(MUXED_ACCOUNT_STARTING_LETTER):
        client_account_id = MuxedAccount.from_account(client_account_id).account_id
    signers = [Ed25519PublicKeySigner(client_account_id, 255)]
    verify_challenge_transaction_signers(
        challenge_transaction,
        server_account_id,
        home_domains,
        web_auth_domain,
        network_passphrase,
        signers,
    )


def _verify_transaction_signatures(
    transaction_envelope: TransactionEnvelope, signers: Sequence[Ed25519PublicKeySigner]
) -> List[Ed25519PublicKeySigner]:
    """Checks if a transaction has been signed by one or more of
    the signers, returning a list of signers that were found to have signed the
    transaction.

    :param transaction_envelope: SEP0010 transaction challenge transaction envelope.
    :param signers: The signers of client account.
    """
    signatures = transaction_envelope.signatures
    if not signatures:
        raise InvalidSep10ChallengeError("Transaction has no signatures.")

    tx_hash = transaction_envelope.hash()

    signers_found = []  # prevent a signature from being reused
    signature_used = set()
    for signer in signers:
        kp = Keypair.from_public_key(signer.account_id)
        for index, decorated_signature in enumerate(transaction_envelope.signatures):
            # Special thanks to Leigh McCulloch for his help
            # See https://github.com/StellarCN/py-stellar-base/issues/252#issuecomment-580882560
            if index in signature_used:
                continue
            if decorated_signature.signature_hint != kp.signature_hint():
                continue
            try:
                kp.verify(tx_hash, decorated_signature.signature)
                signature_used.add(index)
                signers_found.append(signer)
                break
            except BadSignatureError:
                pass
    return signers_found


def _verify_te_signed_by_account_id(
    transaction_envelope: TransactionEnvelope, account_id: str
) -> bool:
    signers_found = _verify_transaction_signatures(
        transaction_envelope, [Ed25519PublicKeySigner(account_id)]
    )
    if not signers_found:
        return False
    return True


def _signer_in_signers(
    signer: Ed25519PublicKeySigner, signers: Sequence[Ed25519PublicKeySigner]
) -> bool:
    for s in signers:
        if s.account_id == signer.account_id:
            return True
    return False



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .envelope_type import EnvelopeType
from .hash_id_preimage_contract_id import HashIDPreimageContractID
from .hash_id_preimage_operation_id import HashIDPreimageOperationID
from .hash_id_preimage_revoke_id import HashIDPreimageRevokeID
from .hash_id_preimage_soroban_authorization import HashIDPreimageSorobanAuthorization

__all__ = ["HashIDPreimage"]


class HashIDPreimage:
    """
    XDR Source Code::

        union HashIDPreimage switch (EnvelopeType type)
        {
        case ENVELOPE_TYPE_OP_ID:
            struct
            {
                AccountID sourceAccount;
                SequenceNumber seqNum;
                uint32 opNum;
            } operationID;
        case ENVELOPE_TYPE_POOL_REVOKE_OP_ID:
            struct
            {
                AccountID sourceAccount;
                SequenceNumber seqNum;
                uint32 opNum;
                PoolID liquidityPoolID;
                Asset asset;
            } revokeID;
        case ENVELOPE_TYPE_CONTRACT_ID:
            struct
            {
                Hash networkID;
                ContractIDPreimage contractIDPreimage;
            } contractID;
        case ENVELOPE_TYPE_SOROBAN_AUTHORIZATION:
            struct
            {
                Hash networkID;
                int64 nonce;
                uint32 signatureExpirationLedger;
                SorobanAuthorizedInvocation invocation;
            } sorobanAuthorization;
        };
    """

    def __init__(
        self,
        type: EnvelopeType,
        operation_id: HashIDPreimageOperationID = None,
        revoke_id: HashIDPreimageRevokeID = None,
        contract_id: HashIDPreimageContractID = None,
        soroban_authorization: HashIDPreimageSorobanAuthorization = None,
    ) -> None:
        self.type = type
        self.operation_id = operation_id
        self.revoke_id = revoke_id
        self.contract_id = contract_id
        self.soroban_authorization = soroban_authorization

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == EnvelopeType.ENVELOPE_TYPE_OP_ID:
            if self.operation_id is None:
                raise ValueError("operation_id should not be None.")
            self.operation_id.pack(packer)
            return
        if self.type == EnvelopeType.ENVELOPE_TYPE_POOL_REVOKE_OP_ID:
            if self.revoke_id is None:
                raise ValueError("revoke_id should not be None.")
            self.revoke_id.pack(packer)
            return
        if self.type == EnvelopeType.ENVELOPE_TYPE_CONTRACT_ID:
            if self.contract_id is None:
                raise ValueError("contract_id should not be None.")
            self.contract_id.pack(packer)
            return
        if self.type == EnvelopeType.ENVELOPE_TYPE_SOROBAN_AUTHORIZATION:
            if self.soroban_authorization is None:
                raise ValueError("soroban_authorization should not be None.")
            self.soroban_authorization.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> HashIDPreimage:
        type = EnvelopeType.unpack(unpacker)
        if type == EnvelopeType.ENVELOPE_TYPE_OP_ID:
            operation_id = HashIDPreimageOperationID.unpack(unpacker)
            return cls(type=type, operation_id=operation_id)
        if type == EnvelopeType.ENVELOPE_TYPE_POOL_REVOKE_OP_ID:
            revoke_id = HashIDPreimageRevokeID.unpack(unpacker)
            return cls(type=type, revoke_id=revoke_id)
        if type == EnvelopeType.ENVELOPE_TYPE_CONTRACT_ID:
            contract_id = HashIDPreimageContractID.unpack(unpacker)
            return cls(type=type, contract_id=contract_id)
        if type == EnvelopeType.ENVELOPE_TYPE_SOROBAN_AUTHORIZATION:
            soroban_authorization = HashIDPreimageSorobanAuthorization.unpack(unpacker)
            return cls(type=type, soroban_authorization=soroban_authorization)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> HashIDPreimage:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> HashIDPreimage:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.operation_id,
                self.revoke_id,
                self.contract_id,
                self.soroban_authorization,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.operation_id == other.operation_id
            and self.revoke_id == other.revoke_id
            and self.contract_id == other.contract_id
            and self.soroban_authorization == other.soroban_authorization
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(
            f"operation_id={self.operation_id}"
        ) if self.operation_id is not None else None
        out.append(
            f"revoke_id={self.revoke_id}"
        ) if self.revoke_id is not None else None
        out.append(
            f"contract_id={self.contract_id}"
        ) if self.contract_id is not None else None
        out.append(
            f"soroban_authorization={self.soroban_authorization}"
        ) if self.soroban_authorization is not None else None
        return f"<HashIDPreimage [{', '.join(out)}]>"



from . import xdr as stellar_xdr

__all__ = ["TimeBounds"]


class TimeBounds:
    """TimeBounds represents the time interval that a transaction is valid.

    The UNIX timestamp (in seconds), determined by ledger time,
    of a lower and upper bound of when this transaction will be valid.
    If a transaction is submitted too early or too late,
    it will fail to make it into the transaction set.
    *max_time* equal ``0`` means that it’s not set.

    See `Stellar's documentation on Transactions
    <https://developers.stellar.org/docs/glossary/transactions/#time-bounds>`__
    for more information on how TimeBounds are used within transactions.


    :param min_time: the UNIX timestamp (in seconds)
    :param max_time: the UNIX timestamp (in seconds)
    :raises: :exc:`ValueError <stellar_sdk.exceptions.ValueError>`: if `max_time` less than `min_time`.
    """

    def __init__(self, min_time: int, max_time: int) -> None:
        if min_time < 0:
            raise ValueError("min_time cannot be negative.")

        if max_time < 0:
            raise ValueError("max_time cannot be negative.")

        if 0 < max_time < min_time:
            raise ValueError("min_time cannot be greater than max_time.")

        self.min_time: int = min_time
        self.max_time: int = max_time

    def to_xdr_object(self) -> stellar_xdr.TimeBounds:
        """Returns the xdr object for this TimeBounds object.

        :return: XDR TimeBounds object
        """
        min_time = stellar_xdr.TimePoint(stellar_xdr.Uint64(self.min_time))
        max_time = stellar_xdr.TimePoint(stellar_xdr.Uint64(self.max_time))
        return stellar_xdr.TimeBounds(min_time, max_time)

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.TimeBounds) -> "TimeBounds":
        """Create a :class:`TimeBounds` from an XDR TimeBounds object.

        :param xdr_object: The XDR TimeBounds object.
        :return: A new :class:`TimeBounds` object from the given XDR TimeBounds object.
        """
        return cls(
            min_time=xdr_object.min_time.time_point.uint64,
            max_time=xdr_object.max_time.time_point.uint64,
        )

    def __hash__(self):
        return hash((self.min_time, self.max_time))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.min_time == other.min_time and self.max_time == other.max_time

    def __str__(self):
        return f"<TimeBounds [min_time={self.min_time}, max_time={self.max_time}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["ContractCostType"]


class ContractCostType(IntEnum):
    """
    XDR Source Code::

        enum ContractCostType {
            // Cost of running 1 wasm instruction
            WasmInsnExec = 0,
            // Cost of allocating a slice of memory (in bytes)
            MemAlloc = 1,
            // Cost of copying a slice of bytes into a pre-allocated memory
            MemCpy = 2,
            // Cost of comparing two slices of memory
            MemCmp = 3,
            // Cost of a host function dispatch, not including the actual work done by
            // the function nor the cost of VM invocation machinary
            DispatchHostFunction = 4,
            // Cost of visiting a host object from the host object storage. Exists to
            // make sure some baseline cost coverage, i.e. repeatly visiting objects
            // by the guest will always incur some charges.
            VisitObject = 5,
            // Cost of serializing an xdr object to bytes
            ValSer = 6,
            // Cost of deserializing an xdr object from bytes
            ValDeser = 7,
            // Cost of computing the sha256 hash from bytes
            ComputeSha256Hash = 8,
            // Cost of computing the ed25519 pubkey from bytes
            ComputeEd25519PubKey = 9,
            // Cost of verifying ed25519 signature of a payload.
            VerifyEd25519Sig = 10,
            // Cost of instantiation a VM from wasm bytes code.
            VmInstantiation = 11,
            // Cost of instantiation a VM from a cached state.
            VmCachedInstantiation = 12,
            // Cost of invoking a function on the VM. If the function is a host function,
            // additional cost will be covered by `DispatchHostFunction`.
            InvokeVmFunction = 13,
            // Cost of computing a keccak256 hash from bytes.
            ComputeKeccak256Hash = 14,
            // Cost of computing an ECDSA secp256k1 signature from bytes.
            ComputeEcdsaSecp256k1Sig = 15,
            // Cost of recovering an ECDSA secp256k1 key from a signature.
            RecoverEcdsaSecp256k1Key = 16,
            // Cost of int256 addition (`+`) and subtraction (`-`) operations
            Int256AddSub = 17,
            // Cost of int256 multiplication (`*`) operation
            Int256Mul = 18,
            // Cost of int256 division (`/`) operation
            Int256Div = 19,
            // Cost of int256 power (`exp`) operation
            Int256Pow = 20,
            // Cost of int256 shift (`shl`, `shr`) operation
            Int256Shift = 21,
            // Cost of drawing random bytes using a ChaCha20 PRNG
            ChaCha20DrawBytes = 22
        };
    """

    WasmInsnExec = 0
    MemAlloc = 1
    MemCpy = 2
    MemCmp = 3
    DispatchHostFunction = 4
    VisitObject = 5
    ValSer = 6
    ValDeser = 7
    ComputeSha256Hash = 8
    ComputeEd25519PubKey = 9
    VerifyEd25519Sig = 10
    VmInstantiation = 11
    VmCachedInstantiation = 12
    InvokeVmFunction = 13
    ComputeKeccak256Hash = 14
    ComputeEcdsaSecp256k1Sig = 15
    RecoverEcdsaSecp256k1Key = 16
    Int256AddSub = 17
    Int256Mul = 18
    Int256Div = 19
    Int256Pow = 20
    Int256Shift = 21
    ChaCha20DrawBytes = 22

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ContractCostType:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ContractCostType:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ContractCostType:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .liquidity_pool_entry_body import LiquidityPoolEntryBody
from .pool_id import PoolID

__all__ = ["LiquidityPoolEntry"]


class LiquidityPoolEntry:
    """
    XDR Source Code::

        struct LiquidityPoolEntry
        {
            PoolID liquidityPoolID;

            union switch (LiquidityPoolType type)
            {
            case LIQUIDITY_POOL_CONSTANT_PRODUCT:
                struct
                {
                    LiquidityPoolConstantProductParameters params;

                    int64 reserveA;        // amount of A in the pool
                    int64 reserveB;        // amount of B in the pool
                    int64 totalPoolShares; // total number of pool shares issued
                    int64 poolSharesTrustLineCount; // number of trust lines for the
                                                    // associated pool shares
                } constantProduct;
            }
            body;
        };
    """

    def __init__(
        self,
        liquidity_pool_id: PoolID,
        body: LiquidityPoolEntryBody,
    ) -> None:
        self.liquidity_pool_id = liquidity_pool_id
        self.body = body

    def pack(self, packer: Packer) -> None:
        self.liquidity_pool_id.pack(packer)
        self.body.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LiquidityPoolEntry:
        liquidity_pool_id = PoolID.unpack(unpacker)
        body = LiquidityPoolEntryBody.unpack(unpacker)
        return cls(
            liquidity_pool_id=liquidity_pool_id,
            body=body,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LiquidityPoolEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LiquidityPoolEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.liquidity_pool_id,
                self.body,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.liquidity_pool_id == other.liquidity_pool_id
            and self.body == other.body
        )

    def __str__(self):
        out = [
            f"liquidity_pool_id={self.liquidity_pool_id}",
            f"body={self.body}",
        ]
        return f"<LiquidityPoolEntry [{', '.join(out)}]>"



from typing import Any, Dict, Generator

from ...call_builder.base.base_call_builder import BaseCallBuilder as _BaseCallBuilder
from ...client.base_sync_client import BaseSyncClient
from ...client.response import Response
from ...exceptions import NotPageableError, raise_request_exception
from ...utils import urljoin_with_query

__all__ = ["BaseCallBuilder"]


class BaseCallBuilder(_BaseCallBuilder):
    """Creates a new :class:`BaseCallBuilder` pointed to server defined by horizon_url.

    This is an **abstract** class. Do not create this object directly, use :class:`stellar_sdk.Server` class.

    :param client: The client instance used to send request.
    :param horizon_url: Horizon server URL.
    """

    def __init__(self, client: BaseSyncClient, **kwargs) -> None:
        super().__init__(**kwargs)
        self.client: BaseSyncClient = client

    def call(self) -> Dict[str, Any]:
        """Triggers a HTTP request using this builder's current configuration.

        :return: If it is called synchronous, the response will be returned. If
            it is called asynchronously, it will return Coroutine.
        :raises:
            | :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`: if you have not successfully
                connected to the server.
            | :exc:`NotFoundError <stellar_sdk.exceptions.NotFoundError>`: if status_code == 404
            | :exc:`BadRequestError <stellar_sdk.exceptions.BadRequestError>`: if 400 <= status_code < 500
                and status_code != 404
            | :exc:`BadResponseError <stellar_sdk.exceptions.BadResponseError>`: if 500 <= status_code < 600
            | :exc:`UnknownRequestError <stellar_sdk.exceptions.UnknownRequestError>`: if an unknown error occurs,
                please submit an issue
        """
        url = urljoin_with_query(self.horizon_url, self.endpoint)
        return self._call(url, self.params)

    def _call(self, url: str, params: dict = None) -> Dict[str, Any]:
        raw_resp = self.client.get(url, params)
        assert isinstance(raw_resp, Response)
        raise_request_exception(raw_resp)
        resp = raw_resp.json()
        self._check_pageable(resp)
        return resp

    def stream(
        self,
    ) -> Generator[Dict[str, Any], None, None]:
        """Creates an EventSource that listens for incoming messages from the server.

        See `Horizon Response Format <https://developers.stellar.org/api/introduction/response-format/>`__

        See `MDN EventSource <https://developer.mozilla.org/en-US/docs/Web/API/EventSource>`__

        :return: an EventSource.

        :raise: :exc:`StreamClientError <stellar_sdk.exceptions.StreamClientError>` - Failed to fetch stream resource.
        """
        url = urljoin_with_query(self.horizon_url, self.endpoint)
        return self.client.stream(url, self.params)

    def next(self) -> Dict[str, Any]:
        if self.next_href is None:
            raise NotPageableError("The next page does not exist.")
        return self._call(self.next_href, None)

    def prev(self) -> Dict[str, Any]:
        if self.prev_href is None:
            raise NotPageableError("The prev page does not exist.")
        return self._call(self.prev_href, None)

    def __hash__(self):
        return hash(
            (
                self.params,
                self.endpoint,
                self.horizon_url,
                self.client,
            )
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.params == other.params
            and self.endpoint == other.endpoint
            and self.horizon_url == other.horizon_url
            and self.client == other.client
        )

    def __str__(self):
        return (
            f"<CallBuilder [horizon_url={self.horizon_url}, "
            f"endpoint={self.endpoint}, "
            f"params={self.params}, "
            f"prev_href={self.prev_href}, "
            f"next_href={self.next_href}, "
            f"client={self.client}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["OperationType"]


class OperationType(IntEnum):
    """
    XDR Source Code::

        enum OperationType
        {
            CREATE_ACCOUNT = 0,
            PAYMENT = 1,
            PATH_PAYMENT_STRICT_RECEIVE = 2,
            MANAGE_SELL_OFFER = 3,
            CREATE_PASSIVE_SELL_OFFER = 4,
            SET_OPTIONS = 5,
            CHANGE_TRUST = 6,
            ALLOW_TRUST = 7,
            ACCOUNT_MERGE = 8,
            INFLATION = 9,
            MANAGE_DATA = 10,
            BUMP_SEQUENCE = 11,
            MANAGE_BUY_OFFER = 12,
            PATH_PAYMENT_STRICT_SEND = 13,
            CREATE_CLAIMABLE_BALANCE = 14,
            CLAIM_CLAIMABLE_BALANCE = 15,
            BEGIN_SPONSORING_FUTURE_RESERVES = 16,
            END_SPONSORING_FUTURE_RESERVES = 17,
            REVOKE_SPONSORSHIP = 18,
            CLAWBACK = 19,
            CLAWBACK_CLAIMABLE_BALANCE = 20,
            SET_TRUST_LINE_FLAGS = 21,
            LIQUIDITY_POOL_DEPOSIT = 22,
            LIQUIDITY_POOL_WITHDRAW = 23,
            INVOKE_HOST_FUNCTION = 24,
            EXTEND_FOOTPRINT_TTL = 25,
            RESTORE_FOOTPRINT = 26
        };
    """

    CREATE_ACCOUNT = 0
    PAYMENT = 1
    PATH_PAYMENT_STRICT_RECEIVE = 2
    MANAGE_SELL_OFFER = 3
    CREATE_PASSIVE_SELL_OFFER = 4
    SET_OPTIONS = 5
    CHANGE_TRUST = 6
    ALLOW_TRUST = 7
    ACCOUNT_MERGE = 8
    INFLATION = 9
    MANAGE_DATA = 10
    BUMP_SEQUENCE = 11
    MANAGE_BUY_OFFER = 12
    PATH_PAYMENT_STRICT_SEND = 13
    CREATE_CLAIMABLE_BALANCE = 14
    CLAIM_CLAIMABLE_BALANCE = 15
    BEGIN_SPONSORING_FUTURE_RESERVES = 16
    END_SPONSORING_FUTURE_RESERVES = 17
    REVOKE_SPONSORSHIP = 18
    CLAWBACK = 19
    CLAWBACK_CLAIMABLE_BALANCE = 20
    SET_TRUST_LINE_FLAGS = 21
    LIQUIDITY_POOL_DEPOSIT = 22
    LIQUIDITY_POOL_WITHDRAW = 23
    INVOKE_HOST_FUNCTION = 24
    EXTEND_FOOTPRINT_TTL = 25
    RESTORE_FOOTPRINT = 26

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> OperationType:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> OperationType:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> OperationType:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .account_id import AccountID
from .asset import Asset
from .uint32 import Uint32

__all__ = ["SetTrustLineFlagsOp"]


class SetTrustLineFlagsOp:
    """
    XDR Source Code::

        struct SetTrustLineFlagsOp
        {
            AccountID trustor;
            Asset asset;

            uint32 clearFlags; // which flags to clear
            uint32 setFlags;   // which flags to set
        };
    """

    def __init__(
        self,
        trustor: AccountID,
        asset: Asset,
        clear_flags: Uint32,
        set_flags: Uint32,
    ) -> None:
        self.trustor = trustor
        self.asset = asset
        self.clear_flags = clear_flags
        self.set_flags = set_flags

    def pack(self, packer: Packer) -> None:
        self.trustor.pack(packer)
        self.asset.pack(packer)
        self.clear_flags.pack(packer)
        self.set_flags.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SetTrustLineFlagsOp:
        trustor = AccountID.unpack(unpacker)
        asset = Asset.unpack(unpacker)
        clear_flags = Uint32.unpack(unpacker)
        set_flags = Uint32.unpack(unpacker)
        return cls(
            trustor=trustor,
            asset=asset,
            clear_flags=clear_flags,
            set_flags=set_flags,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SetTrustLineFlagsOp:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SetTrustLineFlagsOp:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.trustor,
                self.asset,
                self.clear_flags,
                self.set_flags,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.trustor == other.trustor
            and self.asset == other.asset
            and self.clear_flags == other.clear_flags
            and self.set_flags == other.set_flags
        )

    def __str__(self):
        out = [
            f"trustor={self.trustor}",
            f"asset={self.asset}",
            f"clear_flags={self.clear_flags}",
            f"set_flags={self.set_flags}",
        ]
        return f"<SetTrustLineFlagsOp [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .asset import Asset
from .claimable_balance_entry_ext import ClaimableBalanceEntryExt
from .claimable_balance_id import ClaimableBalanceID
from .claimant import Claimant
from .int64 import Int64

__all__ = ["ClaimableBalanceEntry"]


class ClaimableBalanceEntry:
    """
    XDR Source Code::

        struct ClaimableBalanceEntry
        {
            // Unique identifier for this ClaimableBalanceEntry
            ClaimableBalanceID balanceID;

            // List of claimants with associated predicate
            Claimant claimants<10>;

            // Any asset including native
            Asset asset;

            // Amount of asset
            int64 amount;

            // reserved for future use
            union switch (int v)
            {
            case 0:
                void;
            case 1:
                ClaimableBalanceEntryExtensionV1 v1;
            }
            ext;
        };
    """

    def __init__(
        self,
        balance_id: ClaimableBalanceID,
        claimants: List[Claimant],
        asset: Asset,
        amount: Int64,
        ext: ClaimableBalanceEntryExt,
    ) -> None:
        _expect_max_length = 10
        if claimants and len(claimants) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `claimants` should be {_expect_max_length}, but got {len(claimants)}."
            )
        self.balance_id = balance_id
        self.claimants = claimants
        self.asset = asset
        self.amount = amount
        self.ext = ext

    def pack(self, packer: Packer) -> None:
        self.balance_id.pack(packer)
        packer.pack_uint(len(self.claimants))
        for claimants_item in self.claimants:
            claimants_item.pack(packer)
        self.asset.pack(packer)
        self.amount.pack(packer)
        self.ext.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ClaimableBalanceEntry:
        balance_id = ClaimableBalanceID.unpack(unpacker)
        length = unpacker.unpack_uint()
        claimants = []
        for _ in range(length):
            claimants.append(Claimant.unpack(unpacker))
        asset = Asset.unpack(unpacker)
        amount = Int64.unpack(unpacker)
        ext = ClaimableBalanceEntryExt.unpack(unpacker)
        return cls(
            balance_id=balance_id,
            claimants=claimants,
            asset=asset,
            amount=amount,
            ext=ext,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ClaimableBalanceEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ClaimableBalanceEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.balance_id,
                self.claimants,
                self.asset,
                self.amount,
                self.ext,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.balance_id == other.balance_id
            and self.claimants == other.claimants
            and self.asset == other.asset
            and self.amount == other.amount
            and self.ext == other.ext
        )

    def __str__(self):
        out = [
            f"balance_id={self.balance_id}",
            f"claimants={self.claimants}",
            f"asset={self.asset}",
            f"amount={self.amount}",
            f"ext={self.ext}",
        ]
        return f"<ClaimableBalanceEntry [{', '.join(out)}]>"



from ..exceptions import BaseRequestError, SdkError


class StellarTomlNotFoundError(SdkError):
    """If the SEP 0010 toml file not found, the exception will be thrown."""


class InvalidFederationAddress(SdkError):
    """If the federation address is invalid, the exception will be thrown."""


class FederationServerNotFoundError(SdkError):
    """If the federation address is invalid, the exception will be thrown."""


class BadFederationResponseError(BaseRequestError):
    """If the federation address is invalid, the exception will be thrown.

    :param response: client response
    """

    def __init__(self, response) -> None:
        super().__init__(response)
        self.message: str = response.text
        self.status: int = response.status_code


class InvalidSep10ChallengeError(SdkError):
    """If the SEP 0010 validation fails, the exception will be thrown."""


class AccountRequiresMemoError(SdkError):
    """AccountRequiresMemoError is raised when a transaction is trying to submit an
    operation to an account which requires a memo.

    This error contains two attributes to help you identify the account requiring
    the memo and the operation where the account is the destination.

    See `SEP-0029 <https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0029.md>`_ for more
    information.
    """

    def __init__(self, message, account_id, operation_index) -> None:
        super().__init__(message)
        self.account_id: str = account_id
        self.operation_index: int = operation_index



from typing import (
    Any,
    AsyncGenerator,
    Coroutine,
    Dict,
    Generator,
    Mapping,
    Optional,
    Union,
)

__all__ = ["BaseCallBuilder"]


class BaseCallBuilder:
    """Creates a new :class:`BaseCallBuilder` pointed to server defined by horizon_url.

    :param horizon_url: Horizon server URL.
    """

    def __init__(self, horizon_url: str) -> None:
        self.horizon_url: str = horizon_url
        self.params: Dict[str, str] = {}
        self.endpoint: str = ""
        self.prev_href: Optional[str] = None
        self.next_href: Optional[str] = None

    def call(self) -> Union[Dict[str, Any], Coroutine[Any, Any, Dict[str, Any]]]:
        """Triggers a HTTP request using this builder's current configuration.

        :return: If it is called synchronous, the response will be returned. If
            it is called asynchronously, it will return Coroutine.
        :raises:
            | :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`: if you have not successfully
                connected to the server.
            | :exc:`NotFoundError <stellar_sdk.exceptions.NotFoundError>`: if status_code == 404
            | :exc:`BadRequestError <stellar_sdk.exceptions.BadRequestError>`: if 400 <= status_code < 500
                and status_code != 404
            | :exc:`BadResponseError <stellar_sdk.exceptions.BadResponseError>`: if 500 <= status_code < 600
            | :exc:`UnknownRequestError <stellar_sdk.exceptions.UnknownRequestError>`: if an unknown error occurs,
                please submit an issue
        """
        raise NotImplementedError

    def stream(
        self,
    ) -> Union[
        AsyncGenerator[Dict[str, Any], None], Generator[Dict[str, Any], None, None]
    ]:
        """Creates an EventSource that listens for incoming messages from the server.

        See `Horizon Response Format <https://developers.stellar.org/api/introduction/response-format/>`__

        See `MDN EventSource <https://developer.mozilla.org/en-US/docs/Web/API/EventSource>`__

        :return: If it is called synchronous, it will return ``Generator``, If
            it is called asynchronously, it will return ``AsyncGenerator``.

        :raise: :exc:`StreamClientError <stellar_sdk.exceptions.StreamClientError>` - Failed to fetch stream resource.
        """
        raise NotImplementedError

    def next(self):
        raise NotImplementedError

    def prev(self):
        raise NotImplementedError

    def cursor(self, cursor: Union[int, str]):
        """Sets ``cursor`` parameter for the current call. Returns the CallBuilder object on which this method has been called.

        See `Pagination <https://developers.stellar.org/api/introduction/pagination/>`__

        :param cursor: A cursor is a value that points to a specific location in a collection of resources.
        :return: current CallBuilder instance
        """
        self._add_query_param("cursor", cursor)
        return self

    def limit(self, limit: int):
        """Sets ``limit`` parameter for the current call. Returns the CallBuilder object on which this method has been called.

        See `Pagination <https://developers.stellar.org/api/introduction/pagination/>`__

        :param limit: Number of records the server should return.
        :return:
        """
        self._add_query_param("limit", limit)
        return self

    def order(self, desc: bool = True):
        """Sets `order` parameter for the current call. Returns the CallBuilder object on which this method has been called.

        :param desc: Sort direction, ``True`` to get desc sort direction, the default setting is ``True``.
        :return: current CallBuilder instance
        """
        order = "asc"
        if desc:
            order = "desc"
        self._add_query_param("order", order)
        return self

    def _add_query_param(self, key: str, value: Union[str, float, int, bool, None]):
        if value is None:
            pass  # pragma: no cover
        elif value is True:
            self.params[key] = "true"
        elif value is False:
            self.params[key] = "false"
        else:
            self.params[key] = str(value)

    def _check_pageable(self, response: dict) -> None:
        links = response.get("_links")
        if not links:
            return
        prev_page = links.get("prev")
        next_page = links.get("next")
        if prev_page:
            self.prev_href = prev_page.get("href")
        if next_page:
            self.next_href = next_page.get("href")

    def _add_query_params(
        self, params: Mapping[str, Union[str, float, int, bool, None]]
    ) -> None:
        for k, v in params.items():
            self._add_query_param(k, v)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .hash import Hash
from .transaction_signature_payload_tagged_transaction import (
    TransactionSignaturePayloadTaggedTransaction,
)

__all__ = ["TransactionSignaturePayload"]


class TransactionSignaturePayload:
    """
    XDR Source Code::

        struct TransactionSignaturePayload
        {
            Hash networkId;
            union switch (EnvelopeType type)
            {
            // Backwards Compatibility: Use ENVELOPE_TYPE_TX to sign ENVELOPE_TYPE_TX_V0
            case ENVELOPE_TYPE_TX:
                Transaction tx;
            case ENVELOPE_TYPE_TX_FEE_BUMP:
                FeeBumpTransaction feeBump;
            }
            taggedTransaction;
        };
    """

    def __init__(
        self,
        network_id: Hash,
        tagged_transaction: TransactionSignaturePayloadTaggedTransaction,
    ) -> None:
        self.network_id = network_id
        self.tagged_transaction = tagged_transaction

    def pack(self, packer: Packer) -> None:
        self.network_id.pack(packer)
        self.tagged_transaction.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TransactionSignaturePayload:
        network_id = Hash.unpack(unpacker)
        tagged_transaction = TransactionSignaturePayloadTaggedTransaction.unpack(
            unpacker
        )
        return cls(
            network_id=network_id,
            tagged_transaction=tagged_transaction,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TransactionSignaturePayload:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TransactionSignaturePayload:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.network_id,
                self.tagged_transaction,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.network_id == other.network_id
            and self.tagged_transaction == other.tagged_transaction
        )

    def __str__(self):
        out = [
            f"network_id={self.network_id}",
            f"tagged_transaction={self.tagged_transaction}",
        ]
        return f"<TransactionSignaturePayload [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .int64 import Int64
from .uint32 import Uint32

__all__ = ["ConfigSettingContractLedgerCostV0"]


class ConfigSettingContractLedgerCostV0:
    """
    XDR Source Code::

        struct ConfigSettingContractLedgerCostV0
        {
            // Maximum number of ledger entry read operations per ledger
            uint32 ledgerMaxReadLedgerEntries;
            // Maximum number of bytes that can be read per ledger
            uint32 ledgerMaxReadBytes;
            // Maximum number of ledger entry write operations per ledger
            uint32 ledgerMaxWriteLedgerEntries;
            // Maximum number of bytes that can be written per ledger
            uint32 ledgerMaxWriteBytes;

            // Maximum number of ledger entry read operations per transaction
            uint32 txMaxReadLedgerEntries;
            // Maximum number of bytes that can be read per transaction
            uint32 txMaxReadBytes;
            // Maximum number of ledger entry write operations per transaction
            uint32 txMaxWriteLedgerEntries;
            // Maximum number of bytes that can be written per transaction
            uint32 txMaxWriteBytes;

            int64 feeReadLedgerEntry;  // Fee per ledger entry read
            int64 feeWriteLedgerEntry; // Fee per ledger entry write

            int64 feeRead1KB;  // Fee for reading 1KB

            // The following parameters determine the write fee per 1KB.
            // Write fee grows linearly until bucket list reaches this size
            int64 bucketListTargetSizeBytes;
            // Fee per 1KB write when the bucket list is empty
            int64 writeFee1KBBucketListLow;
            // Fee per 1KB write when the bucket list has reached `bucketListTargetSizeBytes`
            int64 writeFee1KBBucketListHigh;
            // Write fee multiplier for any additional data past the first `bucketListTargetSizeBytes`
            uint32 bucketListWriteFeeGrowthFactor;
        };
    """

    def __init__(
        self,
        ledger_max_read_ledger_entries: Uint32,
        ledger_max_read_bytes: Uint32,
        ledger_max_write_ledger_entries: Uint32,
        ledger_max_write_bytes: Uint32,
        tx_max_read_ledger_entries: Uint32,
        tx_max_read_bytes: Uint32,
        tx_max_write_ledger_entries: Uint32,
        tx_max_write_bytes: Uint32,
        fee_read_ledger_entry: Int64,
        fee_write_ledger_entry: Int64,
        fee_read1_kb: Int64,
        bucket_list_target_size_bytes: Int64,
        write_fee1_kb_bucket_list_low: Int64,
        write_fee1_kb_bucket_list_high: Int64,
        bucket_list_write_fee_growth_factor: Uint32,
    ) -> None:
        self.ledger_max_read_ledger_entries = ledger_max_read_ledger_entries
        self.ledger_max_read_bytes = ledger_max_read_bytes
        self.ledger_max_write_ledger_entries = ledger_max_write_ledger_entries
        self.ledger_max_write_bytes = ledger_max_write_bytes
        self.tx_max_read_ledger_entries = tx_max_read_ledger_entries
        self.tx_max_read_bytes = tx_max_read_bytes
        self.tx_max_write_ledger_entries = tx_max_write_ledger_entries
        self.tx_max_write_bytes = tx_max_write_bytes
        self.fee_read_ledger_entry = fee_read_ledger_entry
        self.fee_write_ledger_entry = fee_write_ledger_entry
        self.fee_read1_kb = fee_read1_kb
        self.bucket_list_target_size_bytes = bucket_list_target_size_bytes
        self.write_fee1_kb_bucket_list_low = write_fee1_kb_bucket_list_low
        self.write_fee1_kb_bucket_list_high = write_fee1_kb_bucket_list_high
        self.bucket_list_write_fee_growth_factor = bucket_list_write_fee_growth_factor

    def pack(self, packer: Packer) -> None:
        self.ledger_max_read_ledger_entries.pack(packer)
        self.ledger_max_read_bytes.pack(packer)
        self.ledger_max_write_ledger_entries.pack(packer)
        self.ledger_max_write_bytes.pack(packer)
        self.tx_max_read_ledger_entries.pack(packer)
        self.tx_max_read_bytes.pack(packer)
        self.tx_max_write_ledger_entries.pack(packer)
        self.tx_max_write_bytes.pack(packer)
        self.fee_read_ledger_entry.pack(packer)
        self.fee_write_ledger_entry.pack(packer)
        self.fee_read1_kb.pack(packer)
        self.bucket_list_target_size_bytes.pack(packer)
        self.write_fee1_kb_bucket_list_low.pack(packer)
        self.write_fee1_kb_bucket_list_high.pack(packer)
        self.bucket_list_write_fee_growth_factor.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ConfigSettingContractLedgerCostV0:
        ledger_max_read_ledger_entries = Uint32.unpack(unpacker)
        ledger_max_read_bytes = Uint32.unpack(unpacker)
        ledger_max_write_ledger_entries = Uint32.unpack(unpacker)
        ledger_max_write_bytes = Uint32.unpack(unpacker)
        tx_max_read_ledger_entries = Uint32.unpack(unpacker)
        tx_max_read_bytes = Uint32.unpack(unpacker)
        tx_max_write_ledger_entries = Uint32.unpack(unpacker)
        tx_max_write_bytes = Uint32.unpack(unpacker)
        fee_read_ledger_entry = Int64.unpack(unpacker)
        fee_write_ledger_entry = Int64.unpack(unpacker)
        fee_read1_kb = Int64.unpack(unpacker)
        bucket_list_target_size_bytes = Int64.unpack(unpacker)
        write_fee1_kb_bucket_list_low = Int64.unpack(unpacker)
        write_fee1_kb_bucket_list_high = Int64.unpack(unpacker)
        bucket_list_write_fee_growth_factor = Uint32.unpack(unpacker)
        return cls(
            ledger_max_read_ledger_entries=ledger_max_read_ledger_entries,
            ledger_max_read_bytes=ledger_max_read_bytes,
            ledger_max_write_ledger_entries=ledger_max_write_ledger_entries,
            ledger_max_write_bytes=ledger_max_write_bytes,
            tx_max_read_ledger_entries=tx_max_read_ledger_entries,
            tx_max_read_bytes=tx_max_read_bytes,
            tx_max_write_ledger_entries=tx_max_write_ledger_entries,
            tx_max_write_bytes=tx_max_write_bytes,
            fee_read_ledger_entry=fee_read_ledger_entry,
            fee_write_ledger_entry=fee_write_ledger_entry,
            fee_read1_kb=fee_read1_kb,
            bucket_list_target_size_bytes=bucket_list_target_size_bytes,
            write_fee1_kb_bucket_list_low=write_fee1_kb_bucket_list_low,
            write_fee1_kb_bucket_list_high=write_fee1_kb_bucket_list_high,
            bucket_list_write_fee_growth_factor=bucket_list_write_fee_growth_factor,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ConfigSettingContractLedgerCostV0:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ConfigSettingContractLedgerCostV0:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ledger_max_read_ledger_entries,
                self.ledger_max_read_bytes,
                self.ledger_max_write_ledger_entries,
                self.ledger_max_write_bytes,
                self.tx_max_read_ledger_entries,
                self.tx_max_read_bytes,
                self.tx_max_write_ledger_entries,
                self.tx_max_write_bytes,
                self.fee_read_ledger_entry,
                self.fee_write_ledger_entry,
                self.fee_read1_kb,
                self.bucket_list_target_size_bytes,
                self.write_fee1_kb_bucket_list_low,
                self.write_fee1_kb_bucket_list_high,
                self.bucket_list_write_fee_growth_factor,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ledger_max_read_ledger_entries == other.ledger_max_read_ledger_entries
            and self.ledger_max_read_bytes == other.ledger_max_read_bytes
            and self.ledger_max_write_ledger_entries
            == other.ledger_max_write_ledger_entries
            and self.ledger_max_write_bytes == other.ledger_max_write_bytes
            and self.tx_max_read_ledger_entries == other.tx_max_read_ledger_entries
            and self.tx_max_read_bytes == other.tx_max_read_bytes
            and self.tx_max_write_ledger_entries == other.tx_max_write_ledger_entries
            and self.tx_max_write_bytes == other.tx_max_write_bytes
            and self.fee_read_ledger_entry == other.fee_read_ledger_entry
            and self.fee_write_ledger_entry == other.fee_write_ledger_entry
            and self.fee_read1_kb == other.fee_read1_kb
            and self.bucket_list_target_size_bytes
            == other.bucket_list_target_size_bytes
            and self.write_fee1_kb_bucket_list_low
            == other.write_fee1_kb_bucket_list_low
            and self.write_fee1_kb_bucket_list_high
            == other.write_fee1_kb_bucket_list_high
            and self.bucket_list_write_fee_growth_factor
            == other.bucket_list_write_fee_growth_factor
        )

    def __str__(self):
        out = [
            f"ledger_max_read_ledger_entries={self.ledger_max_read_ledger_entries}",
            f"ledger_max_read_bytes={self.ledger_max_read_bytes}",
            f"ledger_max_write_ledger_entries={self.ledger_max_write_ledger_entries}",
            f"ledger_max_write_bytes={self.ledger_max_write_bytes}",
            f"tx_max_read_ledger_entries={self.tx_max_read_ledger_entries}",
            f"tx_max_read_bytes={self.tx_max_read_bytes}",
            f"tx_max_write_ledger_entries={self.tx_max_write_ledger_entries}",
            f"tx_max_write_bytes={self.tx_max_write_bytes}",
            f"fee_read_ledger_entry={self.fee_read_ledger_entry}",
            f"fee_write_ledger_entry={self.fee_write_ledger_entry}",
            f"fee_read1_kb={self.fee_read1_kb}",
            f"bucket_list_target_size_bytes={self.bucket_list_target_size_bytes}",
            f"write_fee1_kb_bucket_list_low={self.write_fee1_kb_bucket_list_low}",
            f"write_fee1_kb_bucket_list_high={self.write_fee1_kb_bucket_list_high}",
            f"bucket_list_write_fee_growth_factor={self.bucket_list_write_fee_growth_factor}",
        ]
        return f"<ConfigSettingContractLedgerCostV0 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .transaction_result_pair import TransactionResultPair

__all__ = ["TransactionResultSet"]


class TransactionResultSet:
    """
    XDR Source Code::

        struct TransactionResultSet
        {
            TransactionResultPair results<>;
        };
    """

    def __init__(
        self,
        results: List[TransactionResultPair],
    ) -> None:
        _expect_max_length = 4294967295
        if results and len(results) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `results` should be {_expect_max_length}, but got {len(results)}."
            )
        self.results = results

    def pack(self, packer: Packer) -> None:
        packer.pack_uint(len(self.results))
        for results_item in self.results:
            results_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TransactionResultSet:
        length = unpacker.unpack_uint()
        results = []
        for _ in range(length):
            results.append(TransactionResultPair.unpack(unpacker))
        return cls(
            results=results,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TransactionResultSet:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TransactionResultSet:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash((self.results,))

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.results == other.results

    def __str__(self):
        out = [
            f"results={self.results}",
        ]
        return f"<TransactionResultSet [{', '.join(out)}]>"



from json import JSONDecodeError
from typing import Optional

from .client.response import Response

__all__ = [
    "SdkError",
    "BadSignatureError",
    "Ed25519PublicKeyInvalidError",
    "Ed25519SecretSeedInvalidError",
    "MissingEd25519SecretSeedError",
    "MuxedEd25519AccountInvalidError",
    "MemoInvalidException",
    "AssetCodeInvalidError",
    "AssetIssuerInvalidError",
    "NoApproximationError",
    "SignatureExistError",
    "BaseRequestError",
    "BaseHorizonError",
    "ConnectionError",
    "NotFoundError",
    "BadRequestError",
    "BadResponseError",
    "UnknownRequestError",
    "NotPageableError",
    "StreamClientError",
    "FeatureNotEnabledError",
    "SorobanRpcErrorResponse",
    "AccountNotFoundException",
    "PrepareTransactionException",
]

from .soroban_rpc import SimulateTransactionResponse


class SdkError(Exception):
    """Base exception for all stellar sdk related errors"""


class BadSignatureError(SdkError, ValueError):
    """Raised when the signature was forged or otherwise corrupt."""


class Ed25519PublicKeyInvalidError(SdkError, ValueError):
    """Ed25519 public key is incorrect."""


class Ed25519SecretSeedInvalidError(SdkError, ValueError):
    """Ed25519 secret seed is incorrect."""


class MissingEd25519SecretSeedError(SdkError, ValueError):
    """Missing Ed25519 secret seed in the keypair"""


class MuxedEd25519AccountInvalidError(SdkError, ValueError):
    """Muxed Ed25519 public key is incorrect."""


class MemoInvalidException(SdkError, ValueError):
    """Memo is incorrect."""


class AssetCodeInvalidError(SdkError, ValueError):
    """Asset Code is incorrect."""


class AssetIssuerInvalidError(SdkError, ValueError):
    """Asset issuer is incorrect."""


class NoApproximationError(SdkError):
    """Approximation cannot be found"""


class SignatureExistError(SdkError, ValueError):
    """A keypair can only sign a transaction once."""


class BaseRequestError(SdkError):
    """Base class for requests errors."""


class ConnectionError(BaseRequestError):
    """Base class for client connection errors."""


class BaseHorizonError(BaseRequestError):
    """Base class for horizon request errors.

    :param response: client response
    """

    def __init__(self, response: Response) -> None:
        super().__init__(response)

        self.message: str = response.text
        self.status: int = response.status_code

        message = {}
        try:
            message = response.json()
        except JSONDecodeError:
            pass  # pragma: no cover
        self.type: Optional[str] = message.get("type")
        self.title: Optional[str] = message.get("title")
        self.detail: Optional[str] = message.get("detail")
        self.extras: Optional[dict] = message.get("extras")
        self.result_xdr: Optional[str] = message.get("extras", {}).get("result_xdr")

    def __str__(self):
        return self.message


class NotFoundError(BaseHorizonError):
    """This exception is thrown when the requested resource does not exist.
    status_code == 400

    """

    def __init__(self, response):
        super().__init__(response)


class BadRequestError(BaseHorizonError):
    """The request from the client has an error.
    400 <= status_code < 500 and status_code != 404

    """

    def __init__(self, response):
        super().__init__(response)


class BadResponseError(BaseHorizonError):
    """The response from the server has an error.
    500 <= status_code < 600


    """

    def __init__(self, response):
        super().__init__(response)


class UnknownRequestError(BaseHorizonError):
    """Unknown request exception, please create an issue feedback for this issue."""


class NotPageableError(BaseRequestError):
    """There is no previous or next page"""


class StreamClientError(BaseRequestError):
    """Failed to fetch stream resource.

    :param current_cursor: The cursor of the last message obtained can be used for reconnect.
    :param message: error message
    """

    def __init__(self, current_cursor: str, message: str) -> None:
        super().__init__(message)
        self.current_cursor = current_cursor


class FeatureNotEnabledError(SdkError):
    """The feature is not enabled."""


class SorobanRpcErrorResponse(BaseRequestError):
    """The exception is thrown when the RPC server returns an error response."""

    def __init__(
        self, code: int, message: Optional[str], data: Optional[str] = None
    ) -> None:
        super().__init__(message)
        self.code = code
        self.data = data
        self.message = message


class AccountNotFoundException(SdkError):
    """The exception is thrown when trying to load an account that doesn't exist on the Stellar network."""

    def __init__(self, account_id: str) -> None:
        super().__init__(f"Account not found, account_id: {account_id}")
        self.account_id = account_id


class PrepareTransactionException(SdkError):
    """The exception is thrown when trying to prepare a transaction."""

    def __init__(
        self, message: str, simulate_transaction_response: SimulateTransactionResponse
    ) -> None:
        super().__init__(message)
        self.message = message
        self.simulate_transaction_response = simulate_transaction_response


def raise_request_exception(response: Response) -> None:
    status_code = response.status_code
    if status_code == 200:
        pass  # pragma: no cover
    elif status_code == 404:
        raise NotFoundError(response)
    elif 400 <= status_code < 500:
        raise BadRequestError(response)
    elif 500 <= status_code < 600:
        raise BadResponseError(response)
    else:
        raise UnknownRequestError(response)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["SCErrorCode"]


class SCErrorCode(IntEnum):
    """
    XDR Source Code::

        enum SCErrorCode
        {
            SCEC_ARITH_DOMAIN = 0,      // Some arithmetic was undefined (overflow, divide-by-zero).
            SCEC_INDEX_BOUNDS = 1,      // Something was indexed beyond its bounds.
            SCEC_INVALID_INPUT = 2,     // User provided some otherwise-bad data.
            SCEC_MISSING_VALUE = 3,     // Some value was required but not provided.
            SCEC_EXISTING_VALUE = 4,    // Some value was provided where not allowed.
            SCEC_EXCEEDED_LIMIT = 5,    // Some arbitrary limit -- gas or otherwise -- was hit.
            SCEC_INVALID_ACTION = 6,    // Data was valid but action requested was not.
            SCEC_INTERNAL_ERROR = 7,    // The host detected an error in its own logic.
            SCEC_UNEXPECTED_TYPE = 8,   // Some type wasn't as expected.
            SCEC_UNEXPECTED_SIZE = 9    // Something's size wasn't as expected.
        };
    """

    SCEC_ARITH_DOMAIN = 0
    SCEC_INDEX_BOUNDS = 1
    SCEC_INVALID_INPUT = 2
    SCEC_MISSING_VALUE = 3
    SCEC_EXISTING_VALUE = 4
    SCEC_EXCEEDED_LIMIT = 5
    SCEC_INVALID_ACTION = 6
    SCEC_INTERNAL_ERROR = 7
    SCEC_UNEXPECTED_TYPE = 8
    SCEC_UNEXPECTED_SIZE = 9

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCErrorCode:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCErrorCode:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCErrorCode:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



from typing import Union

from ...call_builder.base.base_call_builder import BaseCallBuilder

__all__ = ["BaseTransactionsCallBuilder"]


class BaseTransactionsCallBuilder(BaseCallBuilder):
    """Creates a new :class:`TransactionsCallBuilder` pointed to server defined by horizon_url.

    See `List All Transactions <https://developers.stellar.org/api/resources/transactions/list/>`__ for more information.

    :param horizon_url: Horizon server URL.
    """

    def __init__(self, horizon_url: str) -> None:
        super().__init__(horizon_url)
        self.endpoint: str = "transactions"

    def transaction(self, transaction_hash: str):
        """The transaction details endpoint provides information on a single transaction.
        The transaction hash provided in the hash argument specifies which transaction to load.

        See `Retrieve a Transaction <https://developers.stellar.org/api/resources/transactions/single/>`__ for more information.


        :param transaction_hash: transaction hash
        :return: current TransactionsCallBuilder instance
        """
        self.endpoint = f"transactions/{transaction_hash}"
        return self

    def for_account(self, account_id: str):
        """This endpoint represents all transactions that affected a given account.

        See `Retrieve an Account's Transactions <https://developers.stellar.org/api/resources/accounts/transactions/>`__ for more information.

        :param account_id: account id
        :return: current TransactionsCallBuilder instance
        """
        self.endpoint = f"accounts/{account_id}/transactions"
        return self

    def for_ledger(self, sequence: Union[str, int]):
        """This endpoint represents all transactions in a given ledger.

        See `Retrieve a Ledger's Transactions <https://developers.stellar.org/api/resources/ledgers/transactions/>`__ for more information.

        :param sequence: ledger sequence
        :return: current TransactionsCallBuilder instance
        """
        self.endpoint = f"ledgers/{sequence}/transactions"
        return self

    def for_claimable_balance(self, claimable_balance_id: str):
        """This endpoint represents all transactions referencing a given claimable balance and can be used in streaming mode.

        See `Claimable Balances - Retrieve related Transactions <https://developers.stellar.org/api/resources/claimablebalances/transactions/>`__

        :param claimable_balance_id: This claimable balance’s id encoded in a hex string representation.
        :return: current TransactionsCallBuilder instance
        """
        self.endpoint = f"claimable_balances/{claimable_balance_id}/transactions"
        return self

    def for_liquidity_pool(self, liquidity_pool_id: str):
        """This endpoint represents all transactions referencing a given liquidity pool.

        See `Liquidity Pools - Retrieve related Transactions <https://developers.stellar.org/api/resources/liquiditypools/transactions/>`__

        :param liquidity_pool_id: The ID of the liquidity pool in hex string.
        :return: this TransactionsCallBuilder instance
        """
        self.endpoint = f"liquidity_pools/{liquidity_pool_id}/transactions"
        return self

    def include_failed(self, include_failed: bool):
        """Adds a parameter defining whether to include failed transactions. By default only
        transactions of successful transactions are returned.

        :param include_failed: Set to `True` to include failed transactions.
        :return: current TransactionsCallBuilder instance
        """
        self._add_query_param("include_failed", include_failed)
        return self



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import TYPE_CHECKING

from xdrlib3 import Packer, Unpacker

from .sc_spec_type import SCSpecType

if TYPE_CHECKING:
    from .sc_spec_type_bytes_n import SCSpecTypeBytesN
    from .sc_spec_type_map import SCSpecTypeMap
    from .sc_spec_type_option import SCSpecTypeOption
    from .sc_spec_type_result import SCSpecTypeResult
    from .sc_spec_type_tuple import SCSpecTypeTuple
    from .sc_spec_type_udt import SCSpecTypeUDT
    from .sc_spec_type_vec import SCSpecTypeVec
__all__ = ["SCSpecTypeDef"]


class SCSpecTypeDef:
    """
    XDR Source Code::

        union SCSpecTypeDef switch (SCSpecType type)
        {
        case SC_SPEC_TYPE_VAL:
        case SC_SPEC_TYPE_BOOL:
        case SC_SPEC_TYPE_VOID:
        case SC_SPEC_TYPE_ERROR:
        case SC_SPEC_TYPE_U32:
        case SC_SPEC_TYPE_I32:
        case SC_SPEC_TYPE_U64:
        case SC_SPEC_TYPE_I64:
        case SC_SPEC_TYPE_TIMEPOINT:
        case SC_SPEC_TYPE_DURATION:
        case SC_SPEC_TYPE_U128:
        case SC_SPEC_TYPE_I128:
        case SC_SPEC_TYPE_U256:
        case SC_SPEC_TYPE_I256:
        case SC_SPEC_TYPE_BYTES:
        case SC_SPEC_TYPE_STRING:
        case SC_SPEC_TYPE_SYMBOL:
        case SC_SPEC_TYPE_ADDRESS:
            void;
        case SC_SPEC_TYPE_OPTION:
            SCSpecTypeOption option;
        case SC_SPEC_TYPE_RESULT:
            SCSpecTypeResult result;
        case SC_SPEC_TYPE_VEC:
            SCSpecTypeVec vec;
        case SC_SPEC_TYPE_MAP:
            SCSpecTypeMap map;
        case SC_SPEC_TYPE_TUPLE:
            SCSpecTypeTuple tuple;
        case SC_SPEC_TYPE_BYTES_N:
            SCSpecTypeBytesN bytesN;
        case SC_SPEC_TYPE_UDT:
            SCSpecTypeUDT udt;
        };
    """

    def __init__(
        self,
        type: SCSpecType,
        option: SCSpecTypeOption = None,
        result: SCSpecTypeResult = None,
        vec: SCSpecTypeVec = None,
        map: SCSpecTypeMap = None,
        tuple: SCSpecTypeTuple = None,
        bytes_n: SCSpecTypeBytesN = None,
        udt: SCSpecTypeUDT = None,
    ) -> None:
        self.type = type
        self.option = option
        self.result = result
        self.vec = vec
        self.map = map
        self.tuple = tuple
        self.bytes_n = bytes_n
        self.udt = udt

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == SCSpecType.SC_SPEC_TYPE_VAL:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_BOOL:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_VOID:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_ERROR:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_U32:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_I32:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_U64:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_I64:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_TIMEPOINT:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_DURATION:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_U128:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_I128:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_U256:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_I256:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_BYTES:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_STRING:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_SYMBOL:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_ADDRESS:
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_OPTION:
            if self.option is None:
                raise ValueError("option should not be None.")
            self.option.pack(packer)
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_RESULT:
            if self.result is None:
                raise ValueError("result should not be None.")
            self.result.pack(packer)
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_VEC:
            if self.vec is None:
                raise ValueError("vec should not be None.")
            self.vec.pack(packer)
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_MAP:
            if self.map is None:
                raise ValueError("map should not be None.")
            self.map.pack(packer)
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_TUPLE:
            if self.tuple is None:
                raise ValueError("tuple should not be None.")
            self.tuple.pack(packer)
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_BYTES_N:
            if self.bytes_n is None:
                raise ValueError("bytes_n should not be None.")
            self.bytes_n.pack(packer)
            return
        if self.type == SCSpecType.SC_SPEC_TYPE_UDT:
            if self.udt is None:
                raise ValueError("udt should not be None.")
            self.udt.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCSpecTypeDef:
        type = SCSpecType.unpack(unpacker)
        if type == SCSpecType.SC_SPEC_TYPE_VAL:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_BOOL:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_VOID:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_ERROR:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_U32:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_I32:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_U64:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_I64:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_TIMEPOINT:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_DURATION:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_U128:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_I128:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_U256:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_I256:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_BYTES:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_STRING:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_SYMBOL:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_ADDRESS:
            return cls(type=type)
        if type == SCSpecType.SC_SPEC_TYPE_OPTION:
            from .sc_spec_type_option import SCSpecTypeOption

            option = SCSpecTypeOption.unpack(unpacker)
            return cls(type=type, option=option)
        if type == SCSpecType.SC_SPEC_TYPE_RESULT:
            from .sc_spec_type_result import SCSpecTypeResult

            result = SCSpecTypeResult.unpack(unpacker)
            return cls(type=type, result=result)
        if type == SCSpecType.SC_SPEC_TYPE_VEC:
            from .sc_spec_type_vec import SCSpecTypeVec

            vec = SCSpecTypeVec.unpack(unpacker)
            return cls(type=type, vec=vec)
        if type == SCSpecType.SC_SPEC_TYPE_MAP:
            from .sc_spec_type_map import SCSpecTypeMap

            map = SCSpecTypeMap.unpack(unpacker)
            return cls(type=type, map=map)
        if type == SCSpecType.SC_SPEC_TYPE_TUPLE:
            from .sc_spec_type_tuple import SCSpecTypeTuple

            tuple = SCSpecTypeTuple.unpack(unpacker)
            return cls(type=type, tuple=tuple)
        if type == SCSpecType.SC_SPEC_TYPE_BYTES_N:
            from .sc_spec_type_bytes_n import SCSpecTypeBytesN

            bytes_n = SCSpecTypeBytesN.unpack(unpacker)
            return cls(type=type, bytes_n=bytes_n)
        if type == SCSpecType.SC_SPEC_TYPE_UDT:
            from .sc_spec_type_udt import SCSpecTypeUDT

            udt = SCSpecTypeUDT.unpack(unpacker)
            return cls(type=type, udt=udt)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCSpecTypeDef:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCSpecTypeDef:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.option,
                self.result,
                self.vec,
                self.map,
                self.tuple,
                self.bytes_n,
                self.udt,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.option == other.option
            and self.result == other.result
            and self.vec == other.vec
            and self.map == other.map
            and self.tuple == other.tuple
            and self.bytes_n == other.bytes_n
            and self.udt == other.udt
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(f"option={self.option}") if self.option is not None else None
        out.append(f"result={self.result}") if self.result is not None else None
        out.append(f"vec={self.vec}") if self.vec is not None else None
        out.append(f"map={self.map}") if self.map is not None else None
        out.append(f"tuple={self.tuple}") if self.tuple is not None else None
        out.append(f"bytes_n={self.bytes_n}") if self.bytes_n is not None else None
        out.append(f"udt={self.udt}") if self.udt is not None else None
        return f"<SCSpecTypeDef [{', '.join(out)}]>"
